@@ -21,20 +21,37 @@ use crate::error::Result;
 use crate::input_data::FileId;
 use crate::linker_script::maybe_forced_sysroot;
 use crate::save_dir::SaveDir;
+use crate::timing_phase;
 use hashbrown::HashMap;
 use hashbrown::HashSet;
+use indexmap::IndexSet;
+use itertools::Itertools;
 use jobserver::Acquired;
 use jobserver::Client;
+use object::elf::GNU_PROPERTY_X86_ISA_1_BASELINE;
+use object::elf::GNU_PROPERTY_X86_ISA_1_V2;
+use object::elf::GNU_PROPERTY_X86_ISA_1_V3;
+use object::elf::GNU_PROPERTY_X86_ISA_1_V4;
 use rayon::ThreadPoolBuilder;
 use std::fmt::Display;
+use std::mem::take;
+use std::num::NonZero;
+use std::num::NonZeroU32;
+use std::num::NonZeroU64;
 use std::num::NonZeroUsize;
 use std::path::Path;
 use std::path::PathBuf;
 use std::str::FromStr;
 use std::sync::Arc;
-use std::sync::atomic::AtomicBool;
 use std::sync::atomic::AtomicI64;
-use std::sync::atomic::Ordering;
+
+#[derive(Debug)]
+pub(crate) enum DefsymValue {
+    /// A numeric value (address)
+    Value(u64),
+    /// Reference to another symbol with an optional offset
+    SymbolWithOffset(String, i64),
+}
 
 #[derive(Debug)]
 pub struct Args {
@@ -49,7 +66,7 @@ pub struct Args {
     pub(crate) strip: Strip,
     pub(crate) prepopulate_maps: bool,
     pub(crate) sym_info: Option<String>,
-    pub(crate) merge_strings: bool,
+    pub(crate) merge_sections: bool,
     pub(crate) debug_fuel: Option<AtomicI64>,
     pub(crate) validate_output: bool,
     pub(crate) version_script_path: Option<PathBuf>,
@@ -80,6 +97,12 @@ pub struct Args {
     pub(crate) export_list: Vec<String>,
     pub(crate) export_list_path: Option<PathBuf>,
 
+    /// Symbol definitions from `--defsym` options. Each entry is (symbol_name, value_or_symbol).
+    pub(crate) defsym: Vec<(String, DefsymValue)>,
+
+    /// Section start addresses from `--section-start` options. Maps section name to address.
+    pub(crate) section_start: HashMap<String, u64>,
+
     /// If set, GC stats will be written to the specified filename.
     pub(crate) write_gc_stats: Option<PathBuf>,
 
@@ -102,20 +125,25 @@ pub struct Args {
     pub(crate) got_plt_syms: bool,
     pub(crate) b_symbolic: BSymbolicKind,
     pub(crate) relax: bool,
+    pub(crate) hash_style: HashStyle,
     pub(crate) unresolved_symbols: UnresolvedSymbols,
     pub(crate) error_unresolved_symbols: bool,
     pub(crate) allow_multiple_definitions: bool,
     pub(crate) z_interpose: bool,
+    pub(crate) z_isa: Option<NonZeroU32>,
+    pub(crate) z_stack_size: Option<NonZeroU64>,
+    pub(crate) max_page_size: Option<Alignment>,
 
-    output_kind: Option<OutputKind>,
-    pub(crate) is_dynamic_executable: AtomicBool,
-    relocation_model: RelocationModel,
+    pub(crate) relocation_model: RelocationModel,
+    pub(crate) should_output_executable: bool,
 
     /// The number of actually available threads (considering jobserver)
     pub(crate) available_threads: NonZeroUsize,
 
     pub(crate) numeric_experiments: Vec<Option<u64>>,
 
+    rpath_set: IndexSet<String>,
+
     jobserver_client: Option<Client>,
 }
 
@@ -162,10 +190,20 @@ pub(crate) enum BuildIdOption {
 }
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
-pub(crate) enum OutputKind {
-    StaticExecutable(RelocationModel),
-    DynamicExecutable(RelocationModel),
-    SharedObject,
+pub(crate) enum HashStyle {
+    Gnu,
+    Sysv,
+    Both,
+}
+
+impl HashStyle {
+    pub(crate) const fn includes_gnu(self) -> bool {
+        matches!(self, HashStyle::Gnu | HashStyle::Both)
+    }
+
+    pub(crate) const fn includes_sysv(self) -> bool {
+        matches!(self, HashStyle::Sysv | HashStyle::Both)
+    }
 }
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
@@ -181,7 +219,11 @@ pub(crate) enum Experiment {
     MergeStringSplitParallelism = 0,
 
     /// Number of bytes of string-merge sections before we'll break to a new group.
-    MergeStringMinGroupBytes = 2,
+    MergeStringMinGroupBytes = 1,
+
+    GroupsPerThread = 2,
+
+    MinGroups = 3,
 }
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
@@ -194,6 +236,10 @@ pub(crate) enum FileWriteMode {
     /// update accordingly. If the file is locked due to currently being executed, then our write
     /// will fail.
     UpdateInPlace,
+
+    /// As for `UpdateInPlace`, but if we get an error opening the file for write, fallback to
+    /// unlinking and replacing.
+    UpdateInPlaceWithFallback,
 }
 
 #[derive(Debug, Eq, PartialEq, Clone, Copy)]
@@ -215,8 +261,8 @@ pub struct Modifiers {
 #[derive(Debug, Eq, PartialEq)]
 pub(crate) struct Input {
     pub(crate) spec: InputSpec,
-    /// A directory to search first. Only present when the input came from a linker script, in which
-    /// case this is the directory containing the linker script.
+    /// A directory to search first. Only present when the input came from a linker script, in
+    /// which case this is the directory containing the linker script.
     pub(crate) search_first: Option<PathBuf>,
     pub(crate) modifiers: Modifiers,
 }
@@ -291,7 +337,8 @@ const SILENTLY_IGNORED_FLAGS: &[&str] = &[
 const SILENTLY_IGNORED_SHORT_FLAGS: &[&str] = &[
     "(",
     ")",
-    // On Illumos, the Clang driver inserts a meaningless -C flag before calling any non-GNU ld linker.
+    // On Illumos, the Clang driver inserts a meaningless -C flag before calling any non-GNU ld
+    // linker.
     #[cfg(target_os = "illumos")]
     "C",
 ];
@@ -301,6 +348,8 @@ const IGNORED_FLAGS: &[&str] = &[
     "disable-new-dtags",
     "fix-cortex-a53-835769",
     "fix-cortex-a53-843419",
+    "discard-all",
+    "x", // alias for --discard-all
 ];
 
 // These flags map to the default behavior of the linker.
@@ -326,21 +375,21 @@ impl Default for Args {
             lib_search_path: Vec::new(),
             inputs: Vec::new(),
             output: Arc::from(Path::new("a.out")),
-            is_dynamic_executable: AtomicBool::new(false),
+            should_output_executable: true,
             dynamic_linker: None,
-            output_kind: None,
             time_phase_options: None,
             num_threads: None,
             strip: Strip::Nothing,
-            // For now, we default to --gc-sections. This is different to other linkers, but other than
-            // being different, there doesn't seem to be any downside to doing this. We don't currently do
-            // any less work if we're not GCing sections, but do end up writing more, so --no-gc-sections
-            // will almost always be as slow or slower than --gc-sections. For that reason, the latter is
+            // For now, we default to --gc-sections. This is different to other linkers, but other
+            // than being different, there doesn't seem to be any downside to doing
+            // this. We don't currently do any less work if we're not GCing sections,
+            // but do end up writing more, so --no-gc-sections will almost always be as
+            // slow or slower than --gc-sections. For that reason, the latter is
             // probably a good default.
             gc_sections: true,
             prepopulate_maps: false,
             sym_info: None,
-            merge_strings: true,
+            merge_sections: true,
             copy_relocations: CopyRelocations::Allowed,
             debug_fuel: None,
             validate_output: std::env::var(VALIDATE_ENV).is_ok_and(|v| v == "1"),
@@ -384,15 +433,22 @@ impl Default for Args {
             export_all_dynamic_symbols: false,
             export_list: Vec::new(),
             export_list_path: None,
+            defsym: Vec::new(),
+            section_start: HashMap::new(),
             got_plt_syms: false,
             relax: true,
+            hash_style: HashStyle::Both,
             jobserver_client: None,
             available_threads: NonZeroUsize::new(1).unwrap(),
             unresolved_symbols: UnresolvedSymbols::ReportAll,
             error_unresolved_symbols: true,
             allow_multiple_definitions: false,
             z_interpose: false,
+            z_stack_size: None,
+            z_isa: None,
+            max_page_size: None,
             numeric_experiments: Vec::new(),
+            rpath_set: Default::default(),
         }
     }
 }
@@ -442,11 +498,15 @@ pub(crate) fn parse<F: Fn() -> I, S: AsRef<str>, I: Iterator<Item = S>>(input: F
     }
 
     // Copy relocations are only permitted when building executables.
-    if args.output_kind() == OutputKind::SharedObject {
+    if !args.should_output_executable {
         args.copy_relocations =
             CopyRelocations::Disallowed(CopyRelocationsDisabledReason::SharedObject);
     }
 
+    if !args.rpath_set.is_empty() {
+        args.rpath = Some(take(&mut args.rpath_set).into_iter().join(":"));
+    }
+
     if !args.unrecognized_options.is_empty() {
         let options_list = args.unrecognized_options.join(", ");
         bail!("unrecognized option(s): {}", options_list);
@@ -470,9 +530,13 @@ const fn default_target_arch() -> Architecture {
     {
         Architecture::RISCV64
     }
+    #[cfg(target_arch = "loongarch64")]
+    {
+        Architecture::LoongArch64
+    }
 }
 
-fn read_args_from_file(path: &Path) -> Result<Vec<String>> {
+pub(crate) fn read_args_from_file(path: &Path) -> Result<Vec<String>> {
     let contents = std::fs::read_to_string(path)
         .with_context(|| format!("Failed to read arguments from file `{}`", path.display()))?;
     arguments_from_string(&contents)
@@ -480,17 +544,10 @@ fn read_args_from_file(path: &Path) -> Result<Vec<String>> {
 
 impl Args {
     pub fn parse<F: Fn() -> I, S: AsRef<str>, I: Iterator<Item = S>>(input: F) -> Result<Args> {
+        timing_phase!("Parse args");
         parse(input)
     }
 
-    pub(crate) fn base_address(&self) -> u64 {
-        if self.is_relocatable() {
-            0
-        } else {
-            crate::elf::NON_PIE_START_MEM_ADDRESS
-        }
-    }
-
     /// Uses 1 debug fuel, returning how much fuel remains. Debug fuel is intended to be used when
     /// debugging certain kinds of bugs, so this function isn't normally referenced. To use it, the
     /// caller should take a different branch depending on whether the value is still positive. You
@@ -516,32 +573,6 @@ impl Args {
         }
     }
 
-    pub(crate) fn needs_dynsym(&self) -> bool {
-        self.output_kind().needs_dynsym()
-    }
-
-    pub(crate) fn is_relocatable(&self) -> bool {
-        self.output_kind().is_relocatable()
-    }
-
-    /// Returns whether we need a dynamic section.
-    pub(crate) fn needs_dynamic(&self) -> bool {
-        self.output_kind().needs_dynamic()
-    }
-
-    #[allow(dead_code)]
-    pub(crate) fn should_debug_address(&self, address: u64) -> bool {
-        self.debug_address
-            .is_some_and(|a| address >= a && address < a + 8)
-    }
-
-    pub(crate) fn should_output_symbol_versions(&self) -> bool {
-        matches!(
-            self.output_kind(),
-            OutputKind::DynamicExecutable(_) | OutputKind::SharedObject
-        )
-    }
-
     pub(crate) fn trace_span_for_file(
         &self,
         file_id: FileId,
@@ -554,21 +585,16 @@ impl Args {
         self.should_fork
     }
 
-    pub(crate) fn output_kind(&self) -> OutputKind {
-        self.output_kind.unwrap_or({
-            if self.is_dynamic_executable.load(Ordering::Relaxed) {
-                OutputKind::DynamicExecutable(self.relocation_model)
-            } else {
-                OutputKind::StaticExecutable(self.relocation_model)
-            }
-        })
-    }
-
     pub(crate) fn loadable_segment_alignment(&self) -> Alignment {
+        if let Some(max_page_size) = self.max_page_size {
+            return max_page_size;
+        }
+
         match self.arch {
             Architecture::X86_64 => Alignment { exponent: 12 },
             Architecture::AArch64 => Alignment { exponent: 16 },
             Architecture::RISCV64 => Alignment { exponent: 12 },
+            Architecture::LoongArch64 => Alignment { exponent: 16 },
         }
     }
 
@@ -587,6 +613,8 @@ impl Args {
     ///
     /// <https://www.gnu.org/software/make/manual/html_node/POSIX-Jobserver.html>
     pub fn activate_thread_pool(mut self) -> Result<ActivatedArgs> {
+        timing_phase!("Activate thread pool");
+
         let mut tokens = Vec::new();
         self.available_threads = self.num_threads.unwrap_or_else(|| {
             if let Some(client) = &self.jobserver_client {
@@ -630,19 +658,18 @@ impl Args {
 }
 
 fn parse_number(s: &str) -> Result<u64> {
-    if let Some(s) = s.strip_prefix("0x") {
-        Ok(u64::from_str_radix(s, 16)?)
-    } else {
-        Ok(s.parse::<u64>()?)
-    }
+    crate::parsing::parse_number(s).map_err(|_| crate::error!("Invalid number: {}", s))
 }
 
-fn append_rpath(rpath: &mut Option<String>, rpath_value: &str) {
-    if let Some(rpath) = &mut *rpath {
-        rpath.push(':');
-        rpath.push_str(rpath_value);
-    } else {
-        *rpath = Some(rpath_value.to_string());
+fn parse_defsym_expression(s: &str) -> DefsymValue {
+    use crate::parsing::ParsedSymbolExpression;
+    use crate::parsing::parse_symbol_expression;
+
+    match parse_symbol_expression(s) {
+        ParsedSymbolExpression::Absolute(value) => DefsymValue::Value(value),
+        ParsedSymbolExpression::SymbolWithOffset(sym, offset) => {
+            DefsymValue::SymbolWithOffset(sym.to_owned(), offset)
+        }
     }
 }
 
@@ -657,48 +684,6 @@ impl Default for Modifiers {
     }
 }
 
-impl OutputKind {
-    pub(crate) fn is_executable(self) -> bool {
-        !matches!(self, OutputKind::SharedObject)
-    }
-
-    pub(crate) fn is_shared_object(self) -> bool {
-        matches!(self, OutputKind::SharedObject)
-    }
-
-    pub(crate) fn is_dynamic_executable(self) -> bool {
-        matches!(self, OutputKind::DynamicExecutable(_))
-    }
-
-    pub(crate) fn is_static_executable(self) -> bool {
-        matches!(self, OutputKind::StaticExecutable(_))
-    }
-
-    pub(crate) fn is_relocatable(self) -> bool {
-        matches!(
-            self,
-            OutputKind::StaticExecutable(RelocationModel::Relocatable)
-                | OutputKind::DynamicExecutable(RelocationModel::Relocatable)
-                | OutputKind::SharedObject
-        )
-    }
-
-    pub(crate) fn needs_dynsym(self) -> bool {
-        matches!(
-            self,
-            OutputKind::DynamicExecutable(_)
-                | OutputKind::SharedObject
-                // It seems a bit weird to have dynsym in a static-PIE binary, but that's what GNU
-                // ld does. It just doesn't have any symbols besides the undefined symbol.
-                | OutputKind::StaticExecutable(RelocationModel::Relocatable)
-        )
-    }
-
-    fn needs_dynamic(self) -> bool {
-        self != OutputKind::StaticExecutable(RelocationModel::NonRelocatable)
-    }
-}
-
 /// Parses arguments from a string, handling quoting, escapes etc.
 /// All arguments must be surrounded by a white space.
 fn arguments_from_string(input: &str) -> Result<Vec<String>> {
@@ -818,10 +803,24 @@ struct NoParam;
 struct WithParam;
 struct WithOptionalParam;
 
+#[derive(Clone, Copy)]
+enum SubOptionHandler {
+    /// Handler without value parameter (exact match)
+    NoValue(fn(&mut Args, &mut Vec<Modifiers>) -> Result<()>),
+    /// Handler with value parameter (prefix match)
+    WithValue(fn(&mut Args, &mut Vec<Modifiers>, &str) -> Result<()>),
+}
+
 #[derive(Clone, Copy)]
 struct SubOption {
     help: &'static str,
-    handler: fn(&mut Args, &mut Vec<Modifiers>, &str) -> Result<()>,
+    handler: SubOptionHandler,
+}
+
+impl SubOption {
+    fn with_value(&self) -> bool {
+        matches!(self.handler, SubOptionHandler::WithValue(_))
+    }
 }
 
 impl Default for ArgumentParser {
@@ -965,12 +964,32 @@ impl ArgumentParser {
                     rest.to_owned()
                 };
 
-                // Check if this value corresponds to a registered sub-option
-                if let Some(sub) = handler.sub_options.get(value.as_str()) {
-                    (sub.handler)(args, modifier_stack, &value)?;
+                if let Some((key, param_value)) = value.split_once('=') {
+                    // Value has '=', look up key with trailing '='
+                    if let Some(sub) = handler.sub_options.get(format!("{key}=").as_str()) {
+                        match sub.handler {
+                            SubOptionHandler::NoValue(_) => {
+                                (handler.handler)(args, modifier_stack, &value)?;
+                            }
+                            SubOptionHandler::WithValue(f) => f(args, modifier_stack, param_value)?,
+                        }
+                    } else {
+                        // Fall back to the main handler
+                        (handler.handler)(args, modifier_stack, &value)?;
+                    }
                 } else {
-                    // Fall back to the main handler for unregistered sub-options
-                    (handler.handler)(args, modifier_stack, &value)?;
+                    // No '=' in value, look up exact match
+                    if let Some(sub) = handler.sub_options.get(value.as_str()) {
+                        match sub.handler {
+                            SubOptionHandler::NoValue(f) => f(args, modifier_stack)?,
+                            SubOptionHandler::WithValue(_) => {
+                                bail!("Option -{prefix} {value} requires a value");
+                            }
+                        }
+                    } else {
+                        // Fall back to the main handler
+                        (handler.handler)(args, modifier_stack, &value)?;
+                    }
                 }
                 return Ok(());
             }
@@ -988,7 +1007,7 @@ impl ArgumentParser {
             return Ok(());
         }
 
-        args.save_dir.handle_file(arg)?;
+        args.save_dir.handle_file(arg);
         args.inputs.push(Input {
             spec: InputSpec::File(Box::from(Path::new(arg))),
             search_first: None,
@@ -1050,8 +1069,14 @@ impl ArgumentParser {
                 sub_options.sort_by_key(|(name, _)| *name);
 
                 for (sub_name, sub) in sub_options {
+                    let display_name = if sub.with_value() && sub_name.ends_with('=') {
+                        // sub_name ends with '=' (e.g., "max-page-size="), so add <VALUE>
+                        format!("{sub_name}<VALUE>")
+                    } else {
+                        sub_name.to_string()
+                    };
                     help.push_str(&format!(
-                        "      -{prefix} {sub_name:<30} {sub_help}\n",
+                        "      -{prefix} {display_name:<30} {sub_help}\n",
                         sub_help = sub.help
                     ));
                 }
@@ -1118,12 +1143,35 @@ impl<'a, T> OptionDeclaration<'a, T> {
 
     #[must_use]
     fn sub_option(
+        mut self,
+        name: &'static str,
+        help: &'static str,
+        handler: fn(&mut Args, &mut Vec<Modifiers>) -> Result<()>,
+    ) -> Self {
+        self.sub_options.insert(
+            name,
+            SubOption {
+                help,
+                handler: SubOptionHandler::NoValue(handler),
+            },
+        );
+        self
+    }
+
+    #[must_use]
+    fn sub_option_with_value(
         mut self,
         name: &'static str,
         help: &'static str,
         handler: fn(&mut Args, &mut Vec<Modifiers>, &str) -> Result<()>,
     ) -> Self {
-        self.sub_options.insert(name, SubOption { help, handler });
+        self.sub_options.insert(
+            name,
+            SubOption {
+                help,
+                handler: SubOptionHandler::WithValue(handler),
+            },
+        );
         self
     }
 }
@@ -1221,7 +1269,7 @@ fn setup_argument_parser() -> ArgumentParser {
             };
 
             let dir = handle_sysroot(Path::new(value));
-            args.save_dir.handle_file(value)?;
+            args.save_dir.handle_file(value);
             args.lib_search_path.push(dir);
             Ok(())
         });
@@ -1230,7 +1278,7 @@ fn setup_argument_parser() -> ArgumentParser {
         .declare_with_param()
         .prefix("l")
         .help("Link with library")
-        .sub_option(
+        .sub_option_with_value(
             ":filename",
             "Link with specific file",
             |args, modifier_stack, value| {
@@ -1244,7 +1292,7 @@ fn setup_argument_parser() -> ArgumentParser {
                 Ok(())
             },
         )
-        .sub_option(
+        .sub_option_with_value(
             "libname",
             "Link with library libname.so or libname.a",
             |args, modifier_stack, value| {
@@ -1284,18 +1332,14 @@ fn setup_argument_parser() -> ArgumentParser {
         .declare_with_param()
         .prefix("m")
         .help("Set target architecture")
-        .sub_option(
-            "elf_x86_64",
-            "x86-64 ELF target",
-            |args, _modifier_stack, _value| {
-                args.arch = Architecture::X86_64;
-                Ok(())
-            },
-        )
+        .sub_option("elf_x86_64", "x86-64 ELF target", |args, _| {
+            args.arch = Architecture::X86_64;
+            Ok(())
+        })
         .sub_option(
             "elf_x86_64_sol2",
             "x86-64 ELF target (Solaris)",
-            |args, _modifier_stack, _value| {
+            |args, _| {
                 if args.dynamic_linker.is_none() {
                     args.dynamic_linker = Some(Path::new("/lib/amd64/ld.so.1").into());
                 }
@@ -1303,27 +1347,23 @@ fn setup_argument_parser() -> ArgumentParser {
                 Ok(())
             },
         )
+        .sub_option("aarch64elf", "AArch64 ELF target", |args, _| {
+            args.arch = Architecture::AArch64;
+            Ok(())
+        })
+        .sub_option("aarch64linux", "AArch64 ELF target (Linux)", |args, _| {
+            args.arch = Architecture::AArch64;
+            Ok(())
+        })
+        .sub_option("elf64lriscv", "RISC-V 64-bit ELF target", |args, _| {
+            args.arch = Architecture::RISCV64;
+            Ok(())
+        })
         .sub_option(
-            "aarch64elf",
-            "AArch64 ELF target",
-            |args, _modifier_stack, _value| {
-                args.arch = Architecture::AArch64;
-                Ok(())
-            },
-        )
-        .sub_option(
-            "aarch64linux",
-            "AArch64 ELF target (Linux)",
-            |args, _modifier_stack, _value| {
-                args.arch = Architecture::AArch64;
-                Ok(())
-            },
-        )
-        .sub_option(
-            "elf64lriscv",
-            "RISC-V 64-bit ELF target",
-            |args, _modifier_stack, _value| {
-                args.arch = Architecture::RISCV64;
+            "elf64loongarch",
+            "LoongArch 64-bit ELF target",
+            |args, _| {
+                args.arch = Architecture::LoongArch64;
                 Ok(())
             },
         )
@@ -1335,49 +1375,33 @@ fn setup_argument_parser() -> ArgumentParser {
         .declare_with_param()
         .prefix("z")
         .help("Linker option")
-        .sub_option(
-            "now",
-            "Resolve all symbols immediately",
-            |_args, _modifier_stack, _value| Ok(()),
-        )
+        .sub_option("now", "Resolve all symbols immediately", |_, _| Ok(()))
         .sub_option(
             "origin",
             "Mark object as requiring immediate $ORIGIN",
-            |args, _modifier_stack, _value| {
+            |args, _| {
                 args.needs_origin_handling = true;
                 Ok(())
             },
         )
-        .sub_option(
-            "relro",
-            "Enable RELRO program header",
-            |args, _modifier_stack, _value| {
-                args.relro = true;
-                Ok(())
-            },
-        )
-        .sub_option(
-            "norelro",
-            "Disable RELRO program header",
-            |args, _modifier_stack, _value| {
-                args.relro = false;
-                Ok(())
-            },
-        )
-        .sub_option(
-            "notext",
-            "Do not report DT_TEXTREL as an error",
-            |_args, _modifier_stack, _value| Ok(()),
-        )
-        .sub_option(
-            "nostart-stop-gc",
-            "Disable start/stop symbol GC",
-            |_args, _modifier_stack, _value| Ok(()),
-        )
+        .sub_option("relro", "Enable RELRO program header", |args, _| {
+            args.relro = true;
+            Ok(())
+        })
+        .sub_option("norelro", "Disable RELRO program header", |args, _| {
+            args.relro = false;
+            Ok(())
+        })
+        .sub_option("notext", "Do not report DT_TEXTREL as an error", |_, _| {
+            Ok(())
+        })
+        .sub_option("nostart-stop-gc", "Disable start/stop symbol GC", |_, _| {
+            Ok(())
+        })
         .sub_option(
             "execstack",
             "Mark object as requiring an executable stack",
-            |args, _modifier_stack, _value| {
+            |args, _| {
                 args.execstack = true;
                 Ok(())
             },
@@ -1385,24 +1409,20 @@ fn setup_argument_parser() -> ArgumentParser {
         .sub_option(
             "noexecstack",
             "Mark object as not requiring an executable stack",
-            |args, _modifier_stack, _value| {
+            |args, _| {
                 args.execstack = false;
                 Ok(())
             },
         )
-        .sub_option(
-            "nocopyreloc",
-            "Disable copy relocations",
-            |args, _modifier_stack, _value| {
-                args.copy_relocations =
-                    CopyRelocations::Disallowed(CopyRelocationsDisabledReason::Flag);
-                Ok(())
-            },
-        )
+        .sub_option("nocopyreloc", "Disable copy relocations", |args, _| {
+            args.copy_relocations =
+                CopyRelocations::Disallowed(CopyRelocationsDisabledReason::Flag);
+            Ok(())
+        })
         .sub_option(
             "nodelete",
             "Mark shared object as non-deletable",
-            |args, _modifier_stack, _value| {
+            |args, _| {
                 args.needs_nodelete_handling = true;
                 Ok(())
             },
@@ -1410,7 +1430,7 @@ fn setup_argument_parser() -> ArgumentParser {
         .sub_option(
             "defs",
             "Report unresolved symbol references in object files",
-            |args, _modifier_stack, _value| {
+            |args, _| {
                 args.no_undefined = true;
                 Ok(())
             },
@@ -1418,33 +1438,73 @@ fn setup_argument_parser() -> ArgumentParser {
         .sub_option(
             "undefs",
             "Do not report unresolved symbol references in object files",
-            |args, _modifier_stack, _value| {
+            |args, _| {
                 args.no_undefined = false;
                 Ok(())
             },
         )
-        .sub_option(
-            "muldefs",
-            "Allow multiple definitions",
-            |args, _modifier_stack, _value| {
-                args.allow_multiple_definitions = true;
-                Ok(())
-            },
-        )
-        .sub_option(
-            "lazy",
-            "Use lazy binding (default)",
-            |_args, _modifier_stack, _value| Ok(()),
-        )
+        .sub_option("muldefs", "Allow multiple definitions", |args, _| {
+            args.allow_multiple_definitions = true;
+            Ok(())
+        })
+        .sub_option("lazy", "Use lazy binding (default)", |_, _| Ok(()))
         .sub_option(
             "interpose",
             "Mark object to interpose all DSOs but executable",
-            |args, _modifier_stack, _value| {
+            |args, _| {
                 args.z_interpose = true;
                 Ok(())
             },
         )
-        .execute(|_args, _modifier_stack, _value| Ok(()));
+        .sub_option_with_value(
+            "stack-size=",
+            "Set size of stack segment",
+            |args, _, value| {
+                let size: u64 = parse_number(value)?;
+                args.z_stack_size = NonZero::new(size);
+
+                Ok(())
+            },
+        )
+        .sub_option(
+            "x86-64-baseline",
+            "Mark x86-64-baseline ISA as needed",
+            |args, _| {
+                args.z_isa = NonZero::new(GNU_PROPERTY_X86_ISA_1_BASELINE);
+                Ok(())
+            },
+        )
+        .sub_option("x86-64-v2", "Mark x86-64-v2 ISA as needed", |args, _| {
+            args.z_isa = NonZero::new(GNU_PROPERTY_X86_ISA_1_V2);
+            Ok(())
+        })
+        .sub_option("x86-64-v3", "Mark x86-64-v3 ISA as needed", |args, _| {
+            args.z_isa = NonZero::new(GNU_PROPERTY_X86_ISA_1_V3);
+            Ok(())
+        })
+        .sub_option("x86-64-v4", "Mark x86-64-v4 ISA as needed", |args, _| {
+            args.z_isa = NonZero::new(GNU_PROPERTY_X86_ISA_1_V4);
+            Ok(())
+        })
+        .sub_option_with_value(
+            "max-page-size=",
+            "Set maximum page size for load segments",
+            |args, _, value| {
+                let size: u64 = parse_number(value)?;
+                if !size.is_power_of_two() {
+                    bail!("Invalid alignment {size:#x}");
+                }
+                args.max_page_size = Some(Alignment {
+                    exponent: size.trailing_zeros() as u8,
+                });
+
+                Ok(())
+            },
+        )
+        .execute(|_args, _modifier_stack, value| {
+            warn_unsupported(&("-z ".to_owned() + value))?;
+            Ok(())
+        });
 
     parser
         .declare_with_param()
@@ -1455,7 +1515,7 @@ fn setup_argument_parser() -> ArgumentParser {
                 args.unrecognized_options
                     .push(format!("-R,{value}(filename)"));
             } else {
-                append_rpath(&mut args.rpath, value);
+                args.rpath_set.insert(value.to_string());
             }
             Ok(())
         });
@@ -1540,17 +1600,18 @@ fn setup_argument_parser() -> ArgumentParser {
         .long("Bshareable")
         .help("Create a shared library")
         .execute(|args, _modifier_stack| {
-            args.output_kind = Some(OutputKind::SharedObject);
+            args.should_output_executable = false;
             Ok(())
         });
 
     parser
         .declare()
         .long("pie")
+        .long("pic-executable")
         .help("Create a position-independent executable")
         .execute(|args, _modifier_stack| {
             args.relocation_model = RelocationModel::Relocatable;
-            args.output_kind = None;
+            args.should_output_executable = true;
             Ok(())
         });
 
@@ -1560,7 +1621,7 @@ fn setup_argument_parser() -> ArgumentParser {
         .help("Do not create a position-dependent executable (default)")
         .execute(|args, _modifier_stack| {
             args.relocation_model = RelocationModel::NonRelocatable;
-            args.output_kind = None;
+            args.should_output_executable = true;
             Ok(())
         });
 
@@ -1573,8 +1634,8 @@ fn setup_argument_parser() -> ArgumentParser {
             println!("{}", parser.generate_help());
 
             // The following listing is something autoconf detection relies on.
-            println!("wild: supported targets: elf64-x86-64 elf64-littleaarch64 elf64-littleriscv");
-            println!("wild: supported emulations: elf_x86_64 aarch64elf elf64lriscv");
+            println!("wild: supported targets:elf64 -x86-64 elf64-littleaarch64 elf64-littleriscv elf64-loongarch");
+            println!("wild: supported emulations: elf_x86_64 aarch64elf elf64lriscv elf64loongarch");
 
             std::process::exit(0);
         });
@@ -1774,6 +1835,15 @@ fn setup_argument_parser() -> ArgumentParser {
             Ok(())
         });
 
+    parser
+        .declare()
+        .long("no-eh-frame-hdr")
+        .help("Don't create .eh_frame_hdr section")
+        .execute(|args, _modifier_stack| {
+            args.should_write_eh_frame_hdr = false;
+            Ok(())
+        });
+
     parser
         .declare()
         .long("export-dynamic")
@@ -1808,16 +1878,16 @@ fn setup_argument_parser() -> ArgumentParser {
         .long("rpath")
         .help("Add directory to runtime library search path")
         .execute(|args, _modifier_stack, value| {
-            append_rpath(&mut args.rpath, value);
+            args.rpath_set.insert(value.to_string());
             Ok(())
         });
 
     parser
         .declare()
         .long("no-string-merge")
-        .help("Disable string merging")
+        .help("Disable section merging")
         .execute(|args, _modifier_stack| {
-            args.merge_strings = false;
+            args.merge_sections = false;
             Ok(())
         });
 
@@ -1950,7 +2020,10 @@ fn setup_argument_parser() -> ArgumentParser {
         .help("Exclude libraries")
         .execute(|args, _modifier_stack, value| {
             if value != "ALL" {
+                // For now, just warn and ignore.
+                // FIXME: Support excluding specific libraries.
                 warn_unsupported("--exclude-libs other than ALL")?;
+                return Ok(());
             }
             args.exclude_libs = true;
             Ok(())
@@ -1961,7 +2034,7 @@ fn setup_argument_parser() -> ArgumentParser {
         .long("version-script")
         .help("Use version script")
         .execute(|args, _modifier_stack, value| {
-            args.save_dir.handle_file(value)?;
+            args.save_dir.handle_file(value);
             args.version_script_path = Some(PathBuf::from(value));
             Ok(())
         });
@@ -1972,7 +2045,7 @@ fn setup_argument_parser() -> ArgumentParser {
         .prefix("T")
         .help("Use linker script")
         .execute(|args, _modifier_stack, value| {
-            args.save_dir.handle_file(value)?;
+            args.save_dir.handle_file(value);
             args.add_script(value);
             Ok(())
         });
@@ -2074,14 +2147,57 @@ fn setup_argument_parser() -> ArgumentParser {
             Ok(())
         });
 
+    parser
+        .declare_with_param()
+        .long("defsym")
+        .help("Define a symbol alias: --defsym=symbol=value")
+        .execute(|args, _modifier_stack, value| {
+            let parts: Vec<&str> = value.splitn(2, '=').collect();
+            if parts.len() != 2 {
+                bail!("Invalid --defsym format. Expected: --defsym=symbol=value");
+            }
+            let symbol_name = parts[0].to_owned();
+            let value_str = parts[1];
+
+            let defsym_value = parse_defsym_expression(value_str);
+
+            args.defsym.push((symbol_name, defsym_value));
+            Ok(())
+        });
+
+    parser
+        .declare_with_param()
+        .long("section-start")
+        .help("Set start address for a section: --section-start=.section=address")
+        .execute(|args, _modifier_stack, value| {
+            let parts: Vec<&str> = value.splitn(2, '=').collect();
+            if parts.len() != 2 {
+                bail!("Invalid --section-start format. Expected: --section-start=.section=address");
+            }
+
+            let section_name = parts[0].to_owned();
+            let address = parse_number(parts[1]).with_context(|| {
+                format!(
+                    "Invalid address `{}` in --section-start={}",
+                    parts[1], value
+                )
+            })?;
+            args.section_start.insert(section_name, address);
+
+            Ok(())
+        });
+
     parser
         .declare_with_param()
         .long("hash-style")
         .help("Set hash style")
-        .execute(|_args, _modifier_stack, value| {
-            if value != "gnu" && value != "both" {
-                bail!("Unsupported hash-style `{value}`");
-            }
+        .execute(|args, _modifier_stack, value| {
+            args.hash_style = match value {
+                "gnu" => HashStyle::Gnu,
+                "sysv" => HashStyle::Sysv,
+                "both" => HashStyle::Both,
+                _ => bail!("Unknown hash-style `{value}`"),
+            };
             Ok(())
         });
 
@@ -2153,7 +2269,7 @@ fn setup_argument_parser() -> ArgumentParser {
         .long("sysroot")
         .help("Set system root")
         .execute(|args, _modifier_stack, value| {
-            args.save_dir.handle_file(value)?;
+            args.save_dir.handle_file(value);
             let sysroot = std::fs::canonicalize(value).unwrap_or_else(|_| PathBuf::from(value));
             args.sysroot = Some(Box::from(sysroot.as_path()));
             for path in &mut args.lib_search_path {
@@ -2245,6 +2361,15 @@ fn setup_argument_parser() -> ArgumentParser {
             Ok(())
         });
 
+    parser
+        .declare()
+        .long("no-update-in-place")
+        .help("Delete and recreate the file")
+        .execute(|args, _modifier_stack| {
+            args.file_write_mode = Some(FileWriteMode::UnlinkAndReplace);
+            Ok(())
+        });
+
     parser
         .declare()
         .long("EB")
@@ -2502,6 +2627,11 @@ mod tests {
         "-Rbaz",
         "-R",
         "somewhere",
+        // Adding the same rpath multiple times should not create duplicates
+        "-rpath",
+        "foo/",
+        "-x",
+        "--discard-all",
     ];
 
     const FILE_OPTIONS: &[&str] = &["-pie"];
@@ -2521,7 +2651,6 @@ mod tests {
     }
 
     fn input1_assertions(args: &Args) {
-        assert!(args.is_relocatable());
         assert_eq!(
             args.inputs
                 .iter()
@@ -2557,7 +2686,6 @@ mod tests {
     }
 
     fn inline_and_file_options_assertions(args: &Args) {
-        assert!(args.is_relocatable());
         assert_contains(&args.lib_search_path, "/lib");
     }
 
