@@ -4,10 +4,10 @@ use crate::ensure;
 use crate::error::Context as _;
 use crate::error::Result;
 use crate::resolution::LoadedMetrics;
-use linker_utils::aarch64::DEFAULT_AARCH64_PAGE_IGNORED_MASK;
 use linker_utils::bit_misc::BitExtraction;
 use linker_utils::elf::BitMask;
 use linker_utils::elf::PageMask;
+use linker_utils::elf::RelocationKind;
 use linker_utils::elf::RelocationKindInfo;
 use linker_utils::elf::RelocationSize;
 use linker_utils::elf::SectionType;
@@ -19,7 +19,9 @@ use object::read::elf::CrelIterator;
 use object::read::elf::FileHeader as _;
 use object::read::elf::RelocationSections;
 use object::read::elf::SectionHeader as _;
+use rayon::prelude::*;
 use std::borrow::Cow;
+use std::io::Cursor;
 use std::io::Read as _;
 use std::mem::offset_of;
 use std::ops::Range;
@@ -92,21 +94,11 @@ pub(crate) enum DynamicRelocationSequence<'data> {
 }
 
 pub(crate) trait RelocationSequence<'data> {
-    fn num_relocations(&self) -> usize;
-    fn get_crel(&self, index: usize) -> Crel;
     fn crel_iter(&self) -> impl Iterator<Item = Crel>;
     fn subsequence(&self, range: Range<usize>) -> DynamicRelocationSequence<'data>;
 }
 
 impl<'data> RelocationSequence<'data> for &'data [Rela] {
-    fn num_relocations(&self) -> usize {
-        self.len()
-    }
-
-    fn get_crel(&self, index: usize) -> Crel {
-        Crel::from_rela(&self[index], LittleEndian, false)
-    }
-
     fn crel_iter(&self) -> impl Iterator<Item = Crel> {
         self.iter().map(|r| Crel::from_rela(r, LittleEndian, false))
     }
@@ -117,14 +109,6 @@ impl<'data> RelocationSequence<'data> for &'data [Rela] {
 }
 
 impl RelocationSequence<'static> for Vec<Crel> {
-    fn num_relocations(&self) -> usize {
-        self.len()
-    }
-
-    fn get_crel(&self, index: usize) -> Crel {
-        self[index]
-    }
-
     fn crel_iter(&self) -> impl Iterator<Item = Crel> {
         self.clone().into_iter()
     }
@@ -139,6 +123,9 @@ impl RelocationSequence<'static> for Vec<Crel> {
 const _: () = assert!(!core::mem::needs_drop::<File>());
 
 impl<'data> File<'data> {
+    /// Threshold size for using parallel copy for section data copying.
+    const SECTION_PAR_COPY_SIZE_THRESHOLD: usize = 1_000_000;
+
     pub(crate) fn parse(data: &'data [u8], is_dynamic: bool) -> Result<Self> {
         let header = FileHeader::parse(data)?;
         let endian = header.endian()?;
@@ -249,6 +236,13 @@ impl<'data> File<'data> {
             decompress_into(compression, &data[COMPRESSION_HEADER_SIZE..], out)?;
         } else if section.sh_type(LittleEndian) == object::elf::SHT_NOBITS {
             out.fill(0);
+        } else if data.len() >= Self::SECTION_PAR_COPY_SIZE_THRESHOLD {
+            let threads = rayon::current_num_threads();
+            let chunk_size = (data.len() / threads).max(1);
+
+            data.par_chunks(chunk_size)
+                .zip(out.par_chunks_mut(chunk_size))
+                .for_each(|(src, dst)| dst.copy_from_slice(src));
         } else {
             out.copy_from_slice(data);
         }
@@ -353,10 +347,10 @@ fn decompress_into(
                 flate2::FlushDecompress::Finish,
             )?;
         }
-        // We might use pure Rust implementation for the decompression (ruzstd), however the decompression
-        // speed is not on par with the official C library.
-        // With the official library, the linking time of Clang binary (contains 1GB of debug info sections)
-        // shrinks by 30%!
+        // We might use pure Rust implementation for the decompression (ruzstd), however the
+        // decompression speed is not on par with the official C library.
+        // With the official library, the linking time of Clang binary (contains 1GB of debug info
+        // sections) shrinks by 30%!
         object::elf::ELFCOMPRESS_ZSTD => {
             zstd::stream::Decoder::new(input)?.read_exact(out)?;
         }
@@ -504,18 +498,22 @@ pub(crate) fn get_page_mask(mask: Option<PageMask>) -> PageMaskValue {
     };
 
     match mask {
-        PageMask::SymbolPlusAddendAndPosition => PageMaskValue {
-            symbol_plus_addend: DEFAULT_AARCH64_PAGE_IGNORED_MASK,
-            place: DEFAULT_AARCH64_PAGE_IGNORED_MASK,
+        PageMask::SymbolPlusAddendAndPosition(mask) => PageMaskValue {
+            symbol_plus_addend: !mask,
+            place: !mask,
             ..Default::default()
         },
-        PageMask::GotEntryAndPosition => PageMaskValue {
-            got_entry: DEFAULT_AARCH64_PAGE_IGNORED_MASK,
-            place: DEFAULT_AARCH64_PAGE_IGNORED_MASK,
+        PageMask::GotEntryAndPosition(mask) => PageMaskValue {
+            got_entry: !mask,
+            place: !mask,
             ..Default::default()
         },
-        PageMask::GotBase => PageMaskValue {
-            got: DEFAULT_AARCH64_PAGE_IGNORED_MASK,
+        PageMask::GotBase(mask) => PageMaskValue {
+            got: !mask,
+            ..Default::default()
+        },
+        PageMask::Position(mask) => PageMaskValue {
+            place: !mask,
             ..Default::default()
         },
     }
@@ -529,29 +527,34 @@ pub(crate) fn write_relocation_to_buffer(
 ) -> Result<()> {
     rel_info.verify(value as i64)?;
 
-    match rel_info.size {
-        RelocationSize::ByteSize(byte_size) => {
-            ensure!(
-                byte_size <= output.len(),
-                "Relocation outside of bounds of section"
-            );
-            let value_bytes = value.to_le_bytes();
-            output[..byte_size].copy_from_slice(&value_bytes[..byte_size]);
-        }
-        RelocationSize::BitMasking(BitMask {
-            range,
-            instruction: insn,
-        }) => {
-            let extracted_value = value.extract_bits(range.start..range.end);
-            let negative = (value as i64).is_negative();
-            let output_len = output.len();
-            insn.write_to_value(
-                extracted_value,
-                negative,
-                // We can write a non-text section, so we might need to limit the output buffer slice
-                // e.g. .gcc_except_table is written on riscv64 using the ULEB128 encoding
-                &mut output[..insn.write_windows_size().min(output_len)],
-            );
+    if matches!(rel_info.kind, RelocationKind::PairSubtractionULEB128(..)) {
+        // u64 always fits in 10 bytes in the ULEB format: 64 / 7 = 9.14
+        let mut writer = Cursor::new(vec![0u8; 10]);
+        let n = leb128::write::unsigned(&mut writer, value).expect("Must fit into the buffer");
+        ensure!(
+            output.len() >= n,
+            "cannot write encoded ULEB128 value of {n} bytes"
+        );
+        output[..n].copy_from_slice(&writer.into_inner()[..n]);
+    } else {
+        match rel_info.size {
+            RelocationSize::ByteSize(byte_size) => {
+                ensure!(
+                    byte_size <= output.len(),
+                    "Relocation outside of bounds of section"
+                );
+                let value_bytes = value.to_le_bytes();
+                output[..byte_size].copy_from_slice(&value_bytes[..byte_size]);
+            }
+            RelocationSize::BitMasking(BitMask {
+                range,
+                instruction: insn,
+            }) => {
+                let extracted_value = value.extract_bits(range.start..range.end);
+                let negative = (value as i64).is_negative();
+                let output_len = output.len();
+                insn.write_to_value(extracted_value, negative, &mut output[..output_len]);
+            }
         }
     }
 
