@@ -73,6 +73,7 @@ pub(crate) enum SectionRuleOutcome {
     NoteGnuProperty,
     Debug,
     RiscVAttribute,
+    SortedSection(SectionOutputInfo),
 }
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
@@ -107,7 +108,13 @@ impl<'data> LayoutRulesBuilder<'data> {
         let mut symbol_defs = Vec::new();
 
         for cmd in &input.script.commands {
-            if let linker_script::Command::Sections(sections) = cmd {
+            if let linker_script::Command::SymbolDefinition { name, value } = cmd {
+                let value_str = std::str::from_utf8(value)
+                    .map_err(|_| crate::error!("Invalid UTF-8 in symbol value"))?;
+
+                let placement = crate::parsing::parse_symbol_expression(value_str).to_placement();
+                symbol_defs.push(crate::parsing::InternalSymDefInfo::notype(placement, name));
+            } else if let linker_script::Command::Sections(sections) = cmd {
                 let mut location = None;
 
                 // "Extra alignment" is what we call it when a linker script sets alignment via a
@@ -144,6 +151,7 @@ impl<'data> LayoutRulesBuilder<'data> {
                                             output_sections.add_secondary_section(
                                                 primary_section_id,
                                                 replace(&mut extra_min_alignment, alignment::MIN),
+                                                None,
                                             )
                                         };
 
@@ -284,6 +292,16 @@ impl<'data> SectionRule<'data> {
         )
     }
 
+    const fn prefix_section_sort(
+        name: &'data [u8],
+        section_id: OutputSectionId,
+    ) -> SectionRule<'data> {
+        Self::prefix(
+            name,
+            SectionRuleOutcome::SortedSection(SectionOutputInfo::keep(section_id)),
+        )
+    }
+
     const fn exact(name: &'data [u8], outcome: SectionRuleOutcome) -> SectionRule<'data> {
         SectionRule {
             name,
@@ -325,12 +343,12 @@ const BUILT_IN_RULES: &[SectionRule<'static>] = &[
     ),
     SectionRule::prefix_section(secnames::DATA_SECTION_NAME, output_section_id::DATA),
     SectionRule::prefix_section(secnames::BSS_SECTION_NAME, output_section_id::BSS),
-    SectionRule::prefix_section_keep(
+    SectionRule::prefix_section_sort(
         secnames::INIT_ARRAY_SECTION_NAME,
         output_section_id::INIT_ARRAY,
     ),
     SectionRule::prefix_section_keep(b".ctors", output_section_id::INIT_ARRAY),
-    SectionRule::prefix_section_keep(
+    SectionRule::prefix_section_sort(
         secnames::FINI_ARRAY_SECTION_NAME,
         output_section_id::FINI_ARRAY,
     ),
@@ -342,11 +360,16 @@ const BUILT_IN_RULES: &[SectionRule<'static>] = &[
         output_section_id::GCC_EXCEPT_TABLE,
     ),
     SectionRule::prefix(b".rela", SectionRuleOutcome::Discard),
+    SectionRule::prefix(b".crel", SectionRuleOutcome::Discard),
     SectionRule::exact(secnames::STRTAB_SECTION_NAME, SectionRuleOutcome::Discard),
     SectionRule::exact(secnames::SYMTAB_SECTION_NAME, SectionRuleOutcome::Discard),
     SectionRule::exact(secnames::SHSTRTAB_SECTION_NAME, SectionRuleOutcome::Discard),
     SectionRule::exact(secnames::GROUP_SECTION_NAME, SectionRuleOutcome::Discard),
     SectionRule::exact(secnames::EH_FRAME_SECTION_NAME, SectionRuleOutcome::EhFrame),
+    SectionRule::exact(
+        secnames::SFRAME_SECTION_NAME,
+        SectionRuleOutcome::Section(SectionOutputInfo::keep(output_section_id::SFRAME)),
+    ),
     SectionRule::exact(
         secnames::NOTE_GNU_PROPERTY_SECTION_NAME,
         SectionRuleOutcome::NoteGnuProperty,
