@@ -64,6 +64,10 @@ pub(crate) enum Command<'a> {
     Entry(&'a [u8]),
     #[debug("{}", String::from_utf8_lossy(_0))]
     Version(&'a [u8]),
+    SymbolDefinition {
+        name: &'a [u8],
+        value: &'a [u8],
+    },
 }
 
 #[derive(Debug, PartialEq, Eq)]
@@ -147,7 +151,7 @@ fn parse_token<'input>(input: &mut &'input BStr) -> winnow::Result<&'input [u8]>
 
         Ok(content)
     } else {
-        take_while(1.., |b| !b" (){}\n\t".contains(&b)).parse_next(input)
+        take_while(1.., |b| !b" (){};\n\t".contains(&b)).parse_next(input)
     }
 }
 
@@ -188,7 +192,19 @@ fn parse_command<'input>(input: &mut &'input BStr) -> winnow::Result<Command<'in
         b"SECTIONS" => Command::Sections(parse_sections(input)?),
         b"ENTRY" => Command::Entry(parse_entry(input)?),
         b"VERSION" => Command::Version(parse_version(input)?),
-        other => Command::Arg(other),
+        other => {
+            if input.starts_with(b"=") {
+                // Symbol definition
+                '='.parse_next(input)?;
+                skip_comments_and_whitespace(input)?;
+                let value = take_while(1.., |b| b != b';').parse_next(input)?;
+                let value = value.trim_ascii_end();
+                opt(';').parse_next(input)?;
+                Command::SymbolDefinition { name: other, value }
+            } else {
+                Command::Arg(other)
+            }
+        }
     };
 
     skip_comments_and_whitespace(input)?;
@@ -448,7 +464,7 @@ mod tests {
     use crate::args::InputSpec;
     use itertools::assert_equal;
 
-    fn parse_script<'a>(text: &'a str) -> Result<LinkerScript<'a>> {
+    fn parse_script(text: &str) -> Result<LinkerScript<'_>> {
         LinkerScript::parse(text.as_bytes(), Path::new("test-linker-script.txt"))
     }
 
