@@ -7,9 +7,9 @@
 //!   the multi-part sections.
 //! * Add the section to `test_constant_ids` to make sure the ID is consistent with its position in
 //!   `SECTION_DEFINITIONS`.
-//! * Insert the new section into the output order in `sections_and_segments_events`. The position needs
-//!   to be consistent with the access flags on the section. e.g. if the section is read-only data,
-//!   it should go between the start and end of the read-only segment.
+//! * Insert the new section into the output order in `sections_and_segments_events`. The position
+//!   needs to be consistent with the access flags on the section. e.g. if the section is read-only
+//!   data, it should go between the start and end of the read-only segment.
 //!
 //! Adding a new alignment-base (regular) section is similar to the above, but skip the steps
 //! related to `part_id.rs` and insert later in `SECTION_DEFINITIONS` (probably at the end). Also,
@@ -18,6 +18,7 @@
 use crate::alignment;
 use crate::alignment::Alignment;
 use crate::alignment::NUM_ALIGNMENTS;
+use crate::args::Args;
 use crate::elf;
 use crate::elf::DynamicEntry;
 use crate::elf::GLOBAL_POINTER_SYMBOL_NAME;
@@ -37,6 +38,7 @@ use crate::program_segments::ProgramSegmentId;
 use crate::program_segments::ProgramSegments;
 use crate::program_segments::STACK_SEGMENT_DEF;
 use crate::resolution::SectionSlot;
+use crate::timing_phase;
 use core::slice;
 use hashbrown::HashMap;
 use linker_utils::elf::SectionFlags;
@@ -69,6 +71,14 @@ pub(crate) struct CustomSectionDetails<'data> {
     pub(crate) alignment: Alignment,
 }
 
+#[derive(Clone, Copy, Debug, PartialEq, Eq)]
+pub(crate) struct InitFiniSectionDetail {
+    pub(crate) index: u32,
+    pub(crate) primary: OutputSectionId,
+    pub(crate) priority: u16,
+    pub(crate) alignment: Alignment,
+}
+
 // Single-part sections that we generate ourselves rather than copying directly from input objects.
 pub(crate) const FILE_HEADER: OutputSectionId = part_id::FILE_HEADER.output_section_id();
 pub(crate) const PROGRAM_HEADERS: OutputSectionId = part_id::PROGRAM_HEADERS.output_section_id();
@@ -79,7 +89,9 @@ pub(crate) const GOT: OutputSectionId = part_id::GOT.output_section_id();
 pub(crate) const RELA_PLT: OutputSectionId = part_id::RELA_PLT.output_section_id();
 pub(crate) const EH_FRAME: OutputSectionId = part_id::EH_FRAME.output_section_id();
 pub(crate) const EH_FRAME_HDR: OutputSectionId = part_id::EH_FRAME_HDR.output_section_id();
+pub(crate) const SFRAME: OutputSectionId = part_id::SFRAME.output_section_id();
 pub(crate) const DYNAMIC: OutputSectionId = part_id::DYNAMIC.output_section_id();
+pub(crate) const HASH: OutputSectionId = part_id::SYSV_HASH.output_section_id();
 pub(crate) const GNU_HASH: OutputSectionId = part_id::GNU_HASH.output_section_id();
 pub(crate) const DYNSYM: OutputSectionId = part_id::DYNSYM.output_section_id();
 pub(crate) const DYNSTR: OutputSectionId = part_id::DYNSTR.output_section_id();
@@ -94,6 +106,7 @@ pub(crate) const NOTE_GNU_BUILD_ID: OutputSectionId =
     part_id::NOTE_GNU_BUILD_ID.output_section_id();
 
 pub(crate) const SYMTAB_LOCAL: OutputSectionId = part_id::SYMTAB_LOCAL.output_section_id();
+#[allow(dead_code)]
 pub(crate) const SYMTAB_GLOBAL: OutputSectionId = part_id::SYMTAB_GLOBAL.output_section_id();
 pub(crate) const RELA_DYN_RELATIVE: OutputSectionId =
     part_id::RELA_DYN_RELATIVE.output_section_id();
@@ -131,6 +144,8 @@ pub(crate) struct OutputSections<'data> {
     pub(crate) output_section_indexes: Vec<Option<u16>>,
 
     custom_by_name: HashMap<SectionName<'data>, OutputSectionId>,
+
+    init_fini_by_priority: HashMap<(OutputSectionId, u16), OutputSectionId>,
 }
 
 /// Encodes the order of output sections and the start and end of each program segment. This struct
@@ -155,15 +170,20 @@ struct OutputOrderBuilder<'scope, 'data> {
     active_segment_kinds: Vec<Option<ProgramSegmentId>>,
 
     output_sections: &'scope OutputSections<'data>,
+    secondary: &'scope OutputSectionMap<Vec<OutputSectionId>>,
 }
 
 impl<'scope, 'data> OutputOrderBuilder<'scope, 'data> {
-    fn new(output_sections: &'scope OutputSections<'data>) -> Self {
+    fn new(
+        output_sections: &'scope OutputSections<'data>,
+        secondary: &'scope OutputSectionMap<Vec<OutputSectionId>>,
+    ) -> Self {
         Self {
             events: Vec::new(),
             program_segments: ProgramSegments::empty(),
             output_sections,
             active_segment_kinds: vec![None; PROGRAM_SEGMENT_DEFS.len()],
+            secondary,
         }
     }
 
@@ -175,7 +195,18 @@ impl<'scope, 'data> OutputOrderBuilder<'scope, 'data> {
         }
 
         let section_info = self.output_sections.output_info(section_id);
-        if let Some(location) = section_info.location {
+        debug_assert!(
+            matches!(section_info.kind, SectionKind::Primary(_)),
+            "Attempted to directly emit secondary section {section_id}"
+        );
+
+        // Only emit SetLocation if the section has ALLOC flag, meaning it can be placed
+        // in a segment. Sections without ALLOC (like custom sections before their flags
+        // are propagated) will have their location handled directly in layout_section_parts
+        // via section_info.location.
+        if let Some(location) = section_info.location
+            && section_info.section_flags.contains(shf::ALLOC)
+        {
             self.events.push(OrderEvent::SetLocation(location));
         }
 
@@ -184,6 +215,27 @@ impl<'scope, 'data> OutputOrderBuilder<'scope, 'data> {
         }
 
         self.events.push(OrderEvent::Section(section_id));
+
+        let secondaries: &Vec<OutputSectionId> = self.secondary.get(section_id);
+        // stable ordering: tie-break by original index
+        let mut keyed: Vec<(u16, OutputSectionId)> = secondaries
+            .iter()
+            .map(|&sid| {
+                // default: put non-initfini after all initfini, and keep their relative order
+                let key_pri = match self.output_sections.secondary_order(sid) {
+                    Some(crate::output_section_id::SecondaryOrder::InitFini { priority }) => {
+                        priority
+                    }
+                    None => u16::MAX,
+                };
+                (key_pri, sid)
+            })
+            .collect();
+        keyed.sort_by_key(|(pri, _sid)| *pri);
+
+        for (_pri, sid) in keyed {
+            self.events.push(OrderEvent::Section(sid));
+        }
     }
 
     /// Returns whatever `SegmentStart` and/or `SegmentEnd` events are necessary prior to the start
@@ -366,6 +418,7 @@ pub(crate) struct SectionOutputInfo<'data> {
     pub(crate) min_alignment: Alignment,
     pub(crate) entsize: u64,
     pub(crate) location: Option<linker_script::Location>,
+    pub(crate) secondary_order: Option<SecondaryOrder>,
 }
 
 pub(crate) struct BuiltInSectionDetails {
@@ -375,9 +428,11 @@ pub(crate) struct BuiltInSectionDetails {
     pub(crate) link: &'static [OutputSectionId],
     pub(crate) start_symbol_name: Option<&'static str>,
     pub(crate) end_symbol_name: Option<&'static str>,
+    pub(crate) group_end_symbol_name: Option<&'static str>,
     pub(crate) min_alignment: Alignment,
     info_fn: Option<fn(&InfoInputs) -> u32>,
     pub(crate) keep_if_empty: bool,
+    pub(crate) mark_zero_sized_input_as_content: bool,
     pub(crate) element_size: u64,
     pub(crate) ty: SectionType,
     is_relro: bool,
@@ -390,9 +445,11 @@ const DEFAULT_DEFS: BuiltInSectionDetails = BuiltInSectionDetails {
     link: &[],
     start_symbol_name: None,
     end_symbol_name: None,
+    group_end_symbol_name: None,
     min_alignment: alignment::MIN,
     info_fn: None,
     keep_if_empty: false,
+    mark_zero_sized_input_as_content: true,
     element_size: 0,
     ty: sht::NULL,
     is_relro: false,
@@ -478,6 +535,15 @@ const SECTION_DEFINITIONS: [BuiltInSectionDetails; NUM_BUILT_IN_SECTIONS] = [
         target_segment_type: Some(pt::GNU_EH_FRAME),
         ..DEFAULT_DEFS
     },
+    BuiltInSectionDetails {
+        kind: SectionKind::Primary(SectionName(SFRAME_SECTION_NAME)),
+        ty: sht::GNU_SFRAME,
+        section_flags: shf::ALLOC,
+        min_alignment: alignment::USIZE,
+        target_segment_type: Some(pt::GNU_SFRAME),
+        mark_zero_sized_input_as_content: false,
+        ..DEFAULT_DEFS
+    },
     BuiltInSectionDetails {
         kind: SectionKind::Primary(SectionName(DYNAMIC_SECTION_NAME)),
         ty: sht::DYNAMIC,
@@ -490,6 +556,14 @@ const SECTION_DEFINITIONS: [BuiltInSectionDetails; NUM_BUILT_IN_SECTIONS] = [
         target_segment_type: Some(pt::DYNAMIC),
         ..DEFAULT_DEFS
     },
+    BuiltInSectionDetails {
+        kind: SectionKind::Primary(SectionName(HASH_SECTION_NAME)),
+        ty: sht::HASH,
+        section_flags: shf::ALLOC,
+        link: &[DYNSYM],
+        min_alignment: alignment::SYSV_HASH,
+        ..DEFAULT_DEFS
+    },
     BuiltInSectionDetails {
         kind: SectionKind::Primary(SectionName(GNU_HASH_SECTION_NAME)),
         ty: sht::GNU_HASH,
@@ -554,6 +628,7 @@ const SECTION_DEFINITIONS: [BuiltInSectionDetails; NUM_BUILT_IN_SECTIONS] = [
         ty: sht::NOTE,
         section_flags: shf::ALLOC,
         min_alignment: alignment::NOTE_GNU_PROPERTY,
+        target_segment_type: Some(pt::GNU_PROPERTY),
         ..DEFAULT_DEFS
     },
     BuiltInSectionDetails {
@@ -609,7 +684,8 @@ const SECTION_DEFINITIONS: [BuiltInSectionDetails; NUM_BUILT_IN_SECTIONS] = [
         section_flags: shf::ALLOC.with(shf::WRITE),
         element_size: size_of::<u64>() as u64,
         start_symbol_name: Some("__init_array_start"),
-        end_symbol_name: Some("__init_array_end"),
+        group_end_symbol_name: Some("__init_array_end"),
+        min_alignment: alignment::USIZE,
         is_relro: true,
         ..DEFAULT_DEFS
     },
@@ -619,7 +695,8 @@ const SECTION_DEFINITIONS: [BuiltInSectionDetails; NUM_BUILT_IN_SECTIONS] = [
         section_flags: shf::ALLOC.with(shf::WRITE),
         element_size: size_of::<u64>() as u64,
         start_symbol_name: Some("__fini_array_start"),
-        end_symbol_name: Some("__fini_array_end"),
+        group_end_symbol_name: Some("__fini_array_end"),
+        min_alignment: alignment::USIZE,
         is_relro: true,
         ..DEFAULT_DEFS
     },
@@ -755,6 +832,14 @@ impl OutputSectionId {
             .map_or(alignment::MIN, |d| d.min_alignment)
     }
 
+    pub(crate) fn marks_zero_sized_inputs_as_content(self) -> bool {
+        if let Some(details) = self.opt_built_in_details() {
+            details.mark_zero_sized_input_as_content
+        } else {
+            true
+        }
+    }
+
     pub(crate) fn is_regular(self) -> bool {
         self.0 >= NUM_SINGLE_PART_SECTIONS
     }
@@ -831,20 +916,27 @@ impl Debug for SectionName<'_> {
     }
 }
 
+#[derive(Debug, Clone, Copy)]
+pub(crate) enum SecondaryOrder {
+    InitFini { priority: u16 },
+}
+
 impl CustomSectionIds {
     fn build_output_order_and_program_segments(
         &self,
         output_sections: &OutputSections,
+        secondary: &OutputSectionMap<Vec<OutputSectionId>>,
     ) -> (OutputOrder, ProgramSegments) {
-        let mut builder = OutputOrderBuilder::new(output_sections);
+        let mut builder = OutputOrderBuilder::new(output_sections, secondary);
 
         builder.add_section(FILE_HEADER);
         builder.add_section(PROGRAM_HEADERS);
         builder.add_section(SECTION_HEADERS);
-        builder.add_section(INTERP);
         builder.add_section(NOTE_GNU_PROPERTY);
         builder.add_section(NOTE_GNU_BUILD_ID);
+        builder.add_section(INTERP);
         builder.add_section(NOTE_ABI_TAG);
+        builder.add_section(HASH);
         builder.add_section(GNU_HASH);
         builder.add_section(DYNSYM);
         builder.add_section(DYNSTR);
@@ -852,11 +944,11 @@ impl CustomSectionIds {
         builder.add_section(GNU_VERSION_D);
         builder.add_section(GNU_VERSION_R);
         builder.add_section(RELA_DYN_RELATIVE);
-        builder.add_section(RELA_DYN_GENERAL);
         builder.add_section(RELA_PLT);
         builder.add_section(RODATA);
         builder.add_section(EH_FRAME_HDR);
         builder.add_section(EH_FRAME);
+        builder.add_section(SFRAME);
         builder.add_section(GCC_EXCEPT_TABLE);
         builder.add_sections(&self.ro);
 
@@ -886,7 +978,6 @@ impl CustomSectionIds {
         builder.add_section(RISCV_ATTRIBUTES);
         builder.add_section(SHSTRTAB);
         builder.add_section(SYMTAB_LOCAL);
-        builder.add_section(SYMTAB_GLOBAL);
         builder.add_section(STRTAB);
 
         builder.build()
@@ -894,13 +985,23 @@ impl CustomSectionIds {
 }
 
 impl<'data> OutputSections<'data> {
+    pub(crate) fn secondary_order(&self, id: OutputSectionId) -> Option<SecondaryOrder> {
+        self.section_infos.get(id).secondary_order
+    }
     pub(crate) fn add_sections(
         &mut self,
         custom_sections: &[CustomSectionDetails<'data>],
         sections: &mut [SectionSlot],
+        args: &Args,
     ) {
         for custom in custom_sections {
-            let section_id = self.add_named_section(custom.name, custom.alignment, None);
+            let name_str = std::str::from_utf8(custom.name.bytes()).ok();
+            let location = name_str.and_then(|name| {
+                args.section_start
+                    .get(name)
+                    .map(|&address| linker_script::Location { address })
+            });
+            let section_id = self.add_named_section(custom.name, custom.alignment, location);
 
             if let Some(slot) = sections.get_mut(custom.index.0) {
                 slot.set_part_id(section_id.part_id_with_alignment(custom.alignment));
@@ -924,6 +1025,7 @@ impl<'data> OutputSections<'data> {
                 min_alignment,
                 entsize: 0,
                 location,
+                secondary_order: None,
             })
         })
     }
@@ -932,14 +1034,19 @@ impl<'data> OutputSections<'data> {
         &mut self,
         primary_id: OutputSectionId,
         min_alignment: Alignment,
+        secondary_order: Option<SecondaryOrder>,
     ) -> OutputSectionId {
+        let primary_entsize = self.section_infos.get(primary_id).entsize;
+        let section_flag = self.section_infos.get(primary_id).section_flags;
+        let ty = self.section_infos.get(primary_id).ty;
         self.section_infos.add_new(SectionOutputInfo {
             kind: SectionKind::Secondary(primary_id),
-            section_flags: SectionFlags::empty(),
-            ty: SectionType::from_u32(0),
+            section_flags: section_flag,
+            ty,
             min_alignment,
-            entsize: 0,
+            entsize: primary_entsize,
             location: None,
+            secondary_order,
         })
     }
 
@@ -953,6 +1060,7 @@ impl<'data> OutputSections<'data> {
                 min_alignment: d.min_alignment,
                 entsize: d.element_size,
                 location: None,
+                secondary_order: None,
             })
             .collect();
 
@@ -961,14 +1069,49 @@ impl<'data> OutputSections<'data> {
             base_address,
             custom_by_name: HashMap::new(),
             output_section_indexes: Default::default(),
+            init_fini_by_priority: HashMap::new(),
         }
     }
 
-    #[tracing::instrument(skip_all, name = "Compute output order")]
+    pub(crate) fn bump_min_alignment(&mut self, sid: OutputSectionId, a: Alignment) {
+        let info = self.section_infos.get_mut(sid);
+        info.min_alignment = core::cmp::max(info.min_alignment, a);
+    }
+
+    pub(crate) fn get_or_create_init_fini_secondary(
+        &mut self,
+        primary: OutputSectionId,
+        priority: u16,
+        min_alignment: Alignment,
+    ) -> OutputSectionId {
+        let key = (primary, priority);
+        if let Some(&sid) = self.init_fini_by_priority.get(&key) {
+            self.bump_min_alignment(sid, min_alignment);
+            return sid;
+        }
+
+        let sid = self.add_secondary_section(
+            primary,
+            min_alignment,
+            Some(SecondaryOrder::InitFini { priority }),
+        );
+
+        self.init_fini_by_priority.insert(key, sid);
+        sid
+    }
+
     pub(crate) fn output_order(&self) -> (OutputOrder, ProgramSegments) {
+        timing_phase!("Compute output order");
+
         let mut custom = CustomSectionIds::default();
 
+        let mut secondary: OutputSectionMap<Vec<OutputSectionId>> = self.new_section_map();
+
         self.section_infos.for_each(|id, info| {
+            if let SectionKind::Secondary(primary) = info.kind {
+                secondary.get_mut(primary).push(id);
+                return;
+            }
             if id.as_usize() < NUM_BUILT_IN_SECTIONS {
                 return;
             }
@@ -994,7 +1137,7 @@ impl<'data> OutputSections<'data> {
             }
         });
 
-        custom.build_output_order_and_program_segments(self)
+        custom.build_output_order_and_program_segments(self, &secondary)
     }
 
     #[must_use]
@@ -1051,7 +1194,7 @@ impl<'data> OutputSections<'data> {
                 format!("`{}`", String::from_utf8_lossy(section_name.0))
             }
             SectionKind::Secondary(primary_id) => {
-                format!("`{}` (secondary)", self.display_name(primary_id))
+                format!("{} (secondary)", self.display_name(primary_id))
             }
         }
     }
@@ -1217,6 +1360,7 @@ fn test_constant_ids() {
         (DATA, DATA_SECTION_NAME),
         (EH_FRAME, EH_FRAME_SECTION_NAME),
         (EH_FRAME_HDR, EH_FRAME_HDR_SECTION_NAME),
+        (SFRAME, SFRAME_SECTION_NAME),
         (SHSTRTAB, SHSTRTAB_SECTION_NAME),
         (SYMTAB_LOCAL, SYMTAB_SECTION_NAME),
         (SYMTAB_GLOBAL, &[]),
@@ -1237,6 +1381,7 @@ fn test_constant_ids() {
         (RISCV_ATTRIBUTES, RISCV_ATTRIBUTES_SECTION_NAME),
         (GCC_EXCEPT_TABLE, GCC_EXCEPT_TABLE_SECTION_NAME),
         (INTERP, INTERP_SECTION_NAME),
+        (HASH, HASH_SECTION_NAME),
         (GNU_VERSION, GNU_VERSION_SECTION_NAME),
         (GNU_VERSION_D, GNU_VERSION_D_SECTION_NAME),
         (GNU_VERSION_R, GNU_VERSION_R_SECTION_NAME),
