@@ -1,6 +1,9 @@
 //! Input sections that are marked as string-merge sections need special processing. Our algorithm
 //! is somewhat complicated in an attempt to get good performance. A rough outline of our algorithm
-//! is here with more details throughout the code.
+//! is here with more details throughout the code. Contrary to what the name might suggest, this
+//! algorithm also supports merging non-string sections. The only difference between handling string
+//! and non-string sections is we split the former into multiple slices at the null terminators, and
+//! treat the latter as a single slice.
 //!
 //! We group input sections by the output section into which they are to be placed. We then process
 //! each output section one at a time.
@@ -10,11 +13,11 @@
 //!
 //! With multiple threads, we alternate between two phases:
 //!
-//! Phase 1: We split input sections by looking for null terminators, then we hash the resulting
-//! string and store it in a bucket based on its hash.
+//! Phase 1: We take the whole input sections or split string sections by looking for null
+//! terminators, then we hash the resulting slices and store it in a bucket based on its hash.
 //!
-//! Phase 2: We take the outputs of phase 1 and insert the strings into a hashmap for the bucket
-//! that the string is in. As we do this, we compute bucket-relative offsets for each string and
+//! Phase 2: We take the outputs of phase 1 and insert the slices into a hashmap for the bucket
+//! that the slice is in. As we do this, we compute bucket-relative offsets for each string and
 //! store these into entries in a map that we set up in phase 1.
 //!
 //! Threads can switch between phases multiple times until all work for the section is complete. At
@@ -37,17 +40,23 @@ use crate::part_id::PartId;
 use crate::resolution::ResolvedFile;
 use crate::resolution::ResolvedGroup;
 use crate::resolution::SectionSlot;
+use crate::timing_phase;
+use crate::verbose_timing_phase;
 use crossbeam_queue::ArrayQueue;
 use crossbeam_utils::atomic::AtomicCell;
 use hashbrown::HashMap;
 use itertools::Itertools as _;
+use linker_utils::elf;
+use linker_utils::elf::shf;
 use object::LittleEndian;
 use object::read::elf::Sym as _;
 use rayon::Scope;
 use sharded_offset_map::OffsetMap;
+use sharded_offset_map::ShardedWriter;
 use std::cell::RefCell;
 use std::mem::replace;
 use std::mem::take;
+use std::ops::Range;
 use std::sync::Mutex;
 use std::sync::atomic::AtomicUsize;
 use std::sync::atomic::Ordering;
@@ -57,8 +66,8 @@ use thread_local::ThreadLocal;
 /// splitting parallelism up to the number of threads, but beyond about 24 it doesn't really help.
 const MAX_SPLIT_PARALLELISM: u64 = 24;
 
-/// How large a group of input sections should get before we break to the next group.
-const MIN_GROUP_BYTES: u64 = 140_000;
+/// How large should our chunks of input bytes be.
+const TARGET_GROUP_SIZE_BYTES: u64 = 140_000;
 
 /// Setting this to a higher value increases the potential for parallelism of hash table population
 /// and gives better cache performance. However, it also increases heap allocations. Changing this
@@ -72,6 +81,10 @@ const MERGE_STRING_BUCKETS: usize = 1 << MERGE_STRING_BUCKET_BITS;
 /// spilled to the hashmap.
 const MAP_BLOCK_SIZE: u64 = 256;
 
+pub(crate) struct StringMergeInputs<'data> {
+    input_sections_by_output: OutputSectionMap<Vec<StringMergeInputSection<'data>>>,
+}
+
 #[derive(Debug, Clone, Copy)]
 pub(crate) struct StringMergeSectionSlot {
     pub(crate) part_id: PartId,
@@ -97,6 +110,7 @@ impl StringMergeSectionSlot {
 pub(crate) struct StringMergeSectionExtra<'data> {
     pub(crate) index: object::SectionIndex,
     pub(crate) section_data: &'data [u8],
+    pub(crate) section_flags: elf::SectionFlags,
 }
 
 /// An input offset. We pretend that we've placed all input sections for a given output section one
@@ -112,12 +126,22 @@ impl std::ops::Add<u64> for LinearInputOffset {
     }
 }
 
+impl std::ops::Sub<LinearInputOffset> for LinearInputOffset {
+    type Output = u64;
+
+    fn sub(self, rhs: LinearInputOffset) -> Self::Output {
+        self.0 - rhs.0
+    }
+}
+
 #[derive(Clone, Copy)]
 struct StringMergeInputSection<'data> {
     section_data: &'data [u8],
 
     /// The sum of the sizes of the input sections prior to this one with the same `part_id`.
     start_input_offset: LinearInputOffset,
+
+    is_string: bool,
 }
 
 /// A string from a string-merge section. Includes the null terminator.
@@ -189,14 +213,12 @@ pub(crate) struct MergeStringsSectionBucket<'data> {
 
 /// Merges identical strings from all loaded objects where those strings are from input sections
 /// that are marked with both the SHF_MERGE and SHF_STRINGS flags.
-#[tracing::instrument(skip_all, name = "Merge strings")]
 pub(crate) fn merge_strings<'data>(
-    resolved: &mut [ResolvedGroup<'data>],
+    inputs: &StringMergeInputs<'data>,
     output_sections: &OutputSections,
     args: &Args,
 ) -> Result<OutputSectionMap<MergedStringsSection<'data>>> {
-    let input_sections_by_output =
-        group_merge_string_sections_by_output(resolved, output_sections)?;
+    timing_phase!("Merge strings");
 
     let mut output_string_sections = output_sections.new_section_map::<MergedStringsSection>();
 
@@ -208,17 +230,31 @@ pub(crate) fn merge_strings<'data>(
 
     let reuse_pool = ReusePool::new(MERGE_STRING_BUCKETS * split_parallelism);
 
-    input_sections_by_output.try_for_each(|section_id, input_sections| {
-        let output_section = output_string_sections.get_mut(section_id);
-        output_section.add_input_sections(input_sections, &reuse_pool, args)?;
+    inputs
+        .input_sections_by_output
+        .try_for_each(|section_id, input_sections| {
+            // We later create ArrayQueues with capacity for all input sections and ArrayQueue
+            // panics if asked for zero capacity. Also, spawning tasks and all the other
+            // work we do here would be a waste if we have no input sections.
+            if input_sections.is_empty() {
+                return Ok(());
+            }
 
-        assert_eq!(
-            reuse_pool.available.load(Ordering::Relaxed),
-            reuse_pool.capacity,
-        );
+            verbose_timing_phase!(
+                "Merge section",
+                section_name = output_sections.display_name(section_id)
+            );
 
-        Ok(())
-    })?;
+            let output_section = output_string_sections.get_mut(section_id);
+            output_section.add_input_sections(input_sections, &reuse_pool, args)?;
+
+            assert_eq!(
+                reuse_pool.available.load(Ordering::Relaxed),
+                reuse_pool.capacity,
+            );
+
+            Ok(())
+        })?;
 
     output_string_sections.for_each(|section_id, sec| {
         if sec.len() > 0 {
@@ -240,6 +276,20 @@ pub(crate) fn merge_strings<'data>(
     Ok(output_string_sections)
 }
 
+impl<'data> StringMergeInputs<'data> {
+    pub(crate) fn new(
+        resolved: &mut [ResolvedGroup<'data>],
+        output_sections: &OutputSections,
+    ) -> Result<Self> {
+        Ok(Self {
+            input_sections_by_output: group_merge_string_sections_by_output(
+                resolved,
+                output_sections,
+            )?,
+        })
+    }
+}
+
 // Gather up all the string-merge sections, grouping them by their output section ID. We return a
 // reference to the `MergeStringsFileSection` rather than copying it because it appears to be
 // faster.
@@ -247,6 +297,8 @@ fn group_merge_string_sections_by_output<'data>(
     resolved: &mut [ResolvedGroup<'data>],
     output_sections: &OutputSections,
 ) -> Result<OutputSectionMap<Vec<StringMergeInputSection<'data>>>> {
+    verbose_timing_phase!("Find merge sectionns");
+
     let mut input_sections = output_sections.new_section_map::<Vec<StringMergeInputSection>>();
 
     let mut starting_offsets = output_sections.new_section_map::<LinearInputOffset>();
@@ -256,12 +308,8 @@ fn group_merge_string_sections_by_output<'data>(
             let ResolvedFile::Object(obj) = file else {
                 continue;
             };
-            let Some(non_dynamic) = obj.non_dynamic.as_mut() else {
-                continue;
-            };
-            for extra in &non_dynamic.string_merge_extras {
-                let SectionSlot::MergeStrings(sec) = &mut non_dynamic.sections[extra.index.0]
-                else {
+            for extra in &obj.string_merge_extras {
+                let SectionSlot::MergeStrings(sec) = &mut obj.sections[extra.index.0] else {
                     bail!("Internal error: expected SectionSlot::MergeStrings");
                 };
 
@@ -274,6 +322,7 @@ fn group_merge_string_sections_by_output<'data>(
                     .push(StringMergeInputSection {
                         section_data: extra.section_data,
                         start_input_offset: *starting_offset,
+                        is_string: extra.section_flags.contains(shf::STRINGS),
                     });
 
                 *starting_offset = *starting_offset
@@ -302,6 +351,9 @@ struct SectionGroup<'data, 'offsets, 'sections> {
     index: usize,
     sections: &'sections [StringMergeInputSection<'data>],
     offsets_shard: sharded_offset_map::Shard<'offsets, BucketOffset, MAP_BLOCK_SIZE>,
+
+    /// Restrict to just strings that start within the specified range.
+    range: Range<LinearInputOffset>,
 }
 
 /// Split an input section into strings and hash those strings, collecting the results into
@@ -310,24 +362,63 @@ fn process_input_section<'data, 'offsets>(
     input_section: &StringMergeInputSection<'data>,
     buckets: &mut [Vec<StringToMerge<'data, 'offsets>>; MERGE_STRING_BUCKETS],
     offsets_shard: &mut sharded_offset_map::Shard<'offsets, BucketOffset, MAP_BLOCK_SIZE>,
+    range: &Range<LinearInputOffset>,
 ) -> Result {
     let mut input_offset = input_section.start_input_offset;
     let mut remaining = input_section.section_data;
-    while !remaining.is_empty() {
-        let string = MergeString::take_hashed(&mut remaining)?;
+    if range.start > input_offset {
+        // Non-string merge sections should never be split.
+        debug_assert!(input_section.is_string);
+
+        let offset_in_section = (range.start - input_offset) as usize;
+        let advance = if remaining[offset_in_section - 1] == 0 {
+            // Our range started just after a null character, so we're already at the start of a
+            // string.
+            offset_in_section
+        } else {
+            // Our range start is part way through a string, find end of the string and start from
+            // there.
+            memchr::memchr(0, &remaining[offset_in_section..])
+                .map_or(remaining.len(), |null_offset| {
+                    offset_in_section + null_offset + 1
+                })
+        };
+        input_offset = input_offset + advance as u64;
+        remaining = &remaining[advance..];
+    }
+
+    let mut insert_data = |data: PreHashed<MergeString<'data>>,
+                           input_offset: &mut LinearInputOffset| {
         // Insert 0, then we'll update it later once we know the output offset. We do the
         // initial insertion now since insertions need to happen in sequential order, whereas by
         // the time we know the output offset, we're processing just a single bucket.
+
         let offset_key = match offsets_shard.insert(input_offset.0, BucketOffset(0)) {
             Ok(offset_in_shard) => OffsetOut::InShard(offset_in_shard),
-            Err(_) => OffsetOut::Overflow(input_offset),
+            Err(_) => OffsetOut::Overflow(*input_offset),
         };
-        buckets[(string.hash() as usize) % MERGE_STRING_BUCKETS].push(StringToMerge {
-            string,
+        buckets[(data.hash() as usize) % MERGE_STRING_BUCKETS].push(StringToMerge {
+            string: data,
             offset_out: offset_key,
         });
-        input_offset = input_offset + string.bytes.len() as u64;
+        *input_offset = *input_offset + data.bytes.len() as u64;
+    };
+
+    // Non-string section is just a single slice.
+    if !input_section.is_string {
+        let section_data = MergeString::take_hashed(&mut remaining);
+
+        insert_data(section_data, &mut input_offset);
+        return Ok(());
     }
+
+    // String section, so split at null terminators.
+    while !remaining.is_empty() && input_offset < range.end {
+        let string = MergeString::take_string_hashed(&mut remaining)?;
+
+        insert_data(string, &mut input_offset);
+    }
+
     Ok(())
 }
 
@@ -338,13 +429,6 @@ impl<'data> MergedStringsSection<'data> {
         reuse_pool: &ReusePool,
         args: &Args,
     ) -> Result {
-        // We later create ArrayQueues with capacity for all input sections and ArrayQueue panics if
-        // asked for zero capacity. Also, spawning tasks and all the other work we do here would be
-        // a waste if we have no input sections.
-        if input_sections.is_empty() {
-            return Ok(());
-        }
-
         let mut resources =
             create_split_resources(&mut self.string_offsets, input_sections, reuse_pool, args);
 
@@ -361,14 +445,20 @@ impl<'data> MergedStringsSection<'data> {
             return Err(error);
         }
 
-        // Handle any offsets that didn't fit in their respective blocks in the offset map.
-        let overflow = core::mem::take(&mut resources.overflowed_offsets);
-        overflow
-            .into_iter()
-            .flat_map(|cell| cell.into_inner())
-            .for_each(|o| {
-                self.overflowed_string_offsets.insert(o.input, o.output);
-            });
+        {
+            verbose_timing_phase!("Handle overflows");
+
+            // Handle any offsets that didn't fit in their respective blocks in the offset map.
+            let overflow = core::mem::take(&mut resources.overflowed_offsets);
+            overflow
+                .into_iter()
+                .flat_map(|cell| cell.into_inner())
+                .for_each(|o| {
+                    self.overflowed_string_offsets.insert(o.input, o.output);
+                });
+        }
+
+        verbose_timing_phase!("Finalise merged section");
 
         // Move our buckets out of `resources` and convert it to a regular Vec.
         let mut buckets = resources
@@ -500,59 +590,24 @@ fn create_split_resources<'data, 'offsets, 'scope>(
     reuse_pool: &'scope ReusePool,
     args: &Args,
 ) -> SplitResources<'data, 'offsets, 'scope> {
+    verbose_timing_phase!("Create input section groups");
+
     let input_size = total_input_size(input_sections);
     let mut offset_writer = string_offsets.start_sharded_write(input_size.0);
 
-    let min_group_bytes =
-        args.numeric_experiment(Experiment::MergeStringMinGroupBytes, MIN_GROUP_BYTES);
+    let target_group_size = args
+        .numeric_experiment(
+            Experiment::MergeStringMinGroupBytes,
+            TARGET_GROUP_SIZE_BYTES,
+        )
+        .next_multiple_of(MAP_BLOCK_SIZE) as usize;
 
-    let unprocessed: ArrayQueue<SectionGroup> = ArrayQueue::new(input_sections.len());
+    let groups = split_sections(input_sections, &mut offset_writer, target_group_size);
 
-    let mut group_start_index = 0;
-    let mut group_size = 0;
-
-    input_sections
-        .iter()
-        .enumerate()
-        .for_each(|(index, section)| {
-            let size = (section.section_data.len() as u64).next_multiple_of(MAP_BLOCK_SIZE);
-            let is_last_section = index == input_sections.len() - 1;
-
-            let mut group_end_index = index;
-
-            let new_size = group_size + size;
-            let mut should_end_group = new_size > min_group_bytes
-                && new_size.abs_diff(min_group_bytes) > size.abs_diff(min_group_bytes)
-                && group_size > 0;
-
-            if is_last_section {
-                group_size += size;
-                should_end_group = true;
-                group_end_index += 1;
-            }
-
-            if should_end_group {
-                let first_section = &input_sections[group_start_index];
-                let last_section = &input_sections[group_end_index - 1];
-                let offsets_shard = offset_writer.take_shard(
-                    last_section.start_input_offset.0 - first_section.start_input_offset.0
-                        + (last_section.section_data.len() as u64).next_multiple_of(MAP_BLOCK_SIZE),
-                );
-
-                let r = unprocessed.push(SectionGroup {
-                    index: unprocessed.len(),
-                    sections: &input_sections[group_start_index..group_end_index],
-                    offsets_shard,
-                });
-                // We allocated enough space for each section to be in its own group.
-                assert!(r.is_ok());
-
-                group_size = 0;
-                group_start_index = group_end_index;
-            }
-
-            group_size += size;
-        });
+    let unprocessed: ArrayQueue<SectionGroup> = ArrayQueue::new(groups.len());
+    for group in groups {
+        let _ = unprocessed.push(group);
+    }
 
     let num_groups = unprocessed.len();
     let mut strings_by_bucket_and_group = Vec::new();
@@ -586,6 +641,80 @@ fn create_split_resources<'data, 'offsets, 'scope>(
     resources
 }
 
+/// Split `sections` into slices of at most `size`. A single input section might be split into
+/// multiple groups, or a group might contain multiple input sections. The last slice may be
+/// smaller. If the sections are string sections, then the split will occur after exactly size bytes
+/// unless we run out of sections first. If a section is a non-string merge section, then the whole
+/// section will be taken regardless of size.
+fn split_sections<'data, 'offsets, 'sections>(
+    sections: &'sections [StringMergeInputSection<'data>],
+    offset_writer: &mut ShardedWriter<'offsets, BucketOffset, MAP_BLOCK_SIZE>,
+    size: usize,
+) -> Vec<SectionGroup<'data, 'offsets, 'sections>> {
+    assert!(size.is_multiple_of(MAP_BLOCK_SIZE as usize));
+
+    let mut result = Vec::new();
+
+    let mut section_index = 0;
+    let mut offset_in_section = 0;
+
+    while section_index < sections.len() {
+        // Remaining needs to be signed, since if we encounter non-string merge sections, we'll need
+        // to take the entire section, which may cause us to go negative.
+        let mut remaining = size as isize;
+        let start_section_index = section_index;
+        let first_section_start_offset = offset_in_section;
+        let mut end_section = false;
+
+        // Iterate through sections until we fill `size` bytes
+        while remaining > 0 {
+            let sec = &sections[section_index];
+            let available = (sec.padded_len() - offset_in_section) as isize;
+
+            if available > remaining && sec.is_string {
+                // Still some of this section left for the next group, so don't advance.
+                offset_in_section += remaining as usize;
+                remaining = 0;
+            } else {
+                remaining -= available;
+                if remaining <= 0 || section_index + 1 == sections.len() {
+                    offset_in_section += available as usize;
+                    end_section = true;
+                    break;
+                }
+                section_index += 1;
+                offset_in_section = 0;
+            }
+        }
+
+        let index = result.len();
+        let group_size = size as isize - remaining;
+
+        let linear_start =
+            sections[start_section_index].start_input_offset + first_section_start_offset as u64;
+        let linear_end = sections[section_index].start_input_offset + offset_in_section as u64;
+
+        let offsets_shard = offset_writer.take_shard(group_size as u64);
+
+        debug_assert_eq!(linear_start.0, offsets_shard.base());
+        debug_assert_eq!(linear_end.0, offsets_shard.base() + offsets_shard.len());
+
+        result.push(SectionGroup {
+            sections: &sections[start_section_index..=section_index],
+            range: linear_start..linear_end,
+            index,
+            offsets_shard,
+        });
+
+        if end_section {
+            section_index += 1;
+            offset_in_section = 0;
+        }
+    }
+
+    result
+}
+
 struct ReusePool {
     string_vecs: ArrayQueue<Vec<StringToMerge<'static, 'static>>>,
 
@@ -682,12 +811,19 @@ fn process_input_section_group<'data, 'offsets, 'scope>(
     scope: &Scope<'scope>,
     reservation: &mut PoolReservation,
 ) -> Result {
+    verbose_timing_phase!("Split and hash");
+
     let mut buckets: [Vec<StringToMerge<'data, 'offsets>>; MERGE_STRING_BUCKETS] = [();
         MERGE_STRING_BUCKETS]
         .map(|()| resources.reuse_pool.take_string_merge_vec(reservation));
 
     for section in group_in.sections {
-        process_input_section(section, &mut buckets, &mut group_in.offsets_shard)?;
+        process_input_section(
+            section,
+            &mut buckets,
+            &mut group_in.offsets_shard,
+            &group_in.range,
+        )?;
     }
 
     group_in.offsets_shard.finish();
@@ -713,10 +849,9 @@ fn work_with_bucket<'data, 'scope>(
     resources: &'scope SplitResources<'data, '_, '_>,
     mut bucket: Box<MergeStringsSectionBucket<'data>>,
     scope: &Scope<'scope>,
-) -> Result
-where
-    'scope: 'scope,
-{
+) -> Result {
+    verbose_timing_phase!("Bucket strings");
+
     let mut overflowed_offsets = resources.overflowed_offsets.get_or_default().borrow_mut();
 
     while bucket.next_input_group_index < resources.num_input_groups {
@@ -838,7 +973,9 @@ impl<'data> MergeStringsSectionBucket<'data> {
 impl<'data> MergeString<'data> {
     /// Takes from `source` up to the next null terminator. Returns a prehashed reference to what
     /// was taken.
-    pub(crate) fn take_hashed(source: &mut &'data [u8]) -> Result<PreHashed<MergeString<'data>>> {
+    pub(crate) fn take_string_hashed(
+        source: &mut &'data [u8],
+    ) -> Result<PreHashed<MergeString<'data>>> {
         let len = memchr::memchr(0, source)
             .map(|i| i + 1)
             .context("String in merge-string section is not null-terminated")?;
@@ -847,6 +984,13 @@ impl<'data> MergeString<'data> {
         *source = rest;
         Ok(PreHashed::new(MergeString { bytes }, hash))
     }
+
+    /// Takes the whole `source`. Returns a prehashed reference to what was taken.
+    pub(crate) fn take_hashed(source: &mut &'data [u8]) -> PreHashed<MergeString<'data>> {
+        let bytes = take(source);
+        let hash = crate::hash::hash_bytes(bytes);
+        PreHashed::new(MergeString { bytes }, hash)
+    }
 }
 
 /// Looks for a merged string at `symbol_index` + `addend` in the input and if found, returns its
@@ -945,12 +1089,13 @@ fn find_string(
 }
 
 impl MergedStringStartAddresses {
-    #[tracing::instrument(skip_all, name = "Compute merged string section start addresses")]
     pub(crate) fn compute(
         output_sections: &OutputSections<'_>,
         starting_mem_offsets_by_group: &[OutputSectionPartMap<u64>],
         merge_string_sections: &OutputSectionMap<MergedStringsSection>,
     ) -> Self {
+        timing_phase!("Compute merged string section start addresses");
+
         let mut addresses = output_sections.new_section_map_with(|| [0; MERGE_STRING_BUCKETS]);
         let internal_start_offsets = starting_mem_offsets_by_group.first().unwrap();
         merge_string_sections.for_each(|section_id, sec| {
@@ -971,6 +1116,15 @@ impl MergedStringStartAddresses {
     }
 }
 
+impl StringMergeInputSection<'_> {
+    /// Returns the length of this section's data rounded up to the next multiple of the block size.
+    fn padded_len(&self) -> usize {
+        self.section_data
+            .len()
+            .next_multiple_of(MAP_BLOCK_SIZE as usize)
+    }
+}
+
 impl std::fmt::Display for MergeString<'_> {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         write!(f, "{}", String::from_utf8_lossy(self.bytes))
