@@ -12,6 +12,7 @@ use crate::hash::PreHashed;
 use crate::input_data::ScriptData;
 use crate::linker_script::skip_comments_and_whitespace;
 use crate::symbol::UnversionedSymbolName;
+use crate::timing_phase;
 use glob::Pattern;
 use hashbrown::HashMap;
 use hashbrown::HashSet;
@@ -24,32 +25,55 @@ use winnow::error::FromExternalError;
 use winnow::token::take_until;
 use winnow::token::take_while;
 
-#[derive(Debug, Default)]
+#[derive(Debug, Default, PartialEq, Eq)]
 pub(crate) struct MatchRules<'data> {
     pub(crate) general: BasicMatchRules<'data>,
     pub(crate) cxx: BasicMatchRules<'data>,
 }
 
-#[derive(Debug, Default)]
+#[derive(Debug, Default, PartialEq, Eq)]
 pub(crate) struct VersionBody<'data> {
     globals: MatchRules<'data>,
     locals: MatchRules<'data>,
 }
 
-#[derive(Debug, Default)]
+#[derive(Debug, Default, PartialEq, Eq)]
 pub(crate) struct Version<'data> {
     pub(crate) name: &'data [u8],
     pub(crate) parent_index: Option<u16>,
     pub(crate) version_body: VersionBody<'data>,
 }
 
-/// A version script. See https://sourceware.org/binutils/docs/ld/VERSION.html
-#[derive(Debug, Default)]
-pub(crate) struct VersionScript<'data> {
+/// A general version script. See https://sourceware.org/binutils/docs/ld/VERSION.html
+#[derive(Debug, Default, PartialEq, Eq)]
+pub(crate) struct RegularVersionScript<'data> {
     versions: Vec<Version<'data>>,
     version_name_mapping: HashMap<&'data [u8], usize>,
 }
 
+/// An optimized version script for Rustc.
+/// It declares all symbols as local except for the explicitly listed global symbols.
+/// Only contains general (non-C++) exact symbol matchers.
+/// Doesn't use actual versioning.
+#[derive(Debug, Default, PartialEq, Eq)]
+pub(crate) struct RustVersionScript<'data> {
+    pub(crate) global: Vec<&'data [u8]>,
+}
+
+#[derive(Debug, PartialEq, Eq)]
+/// Possibly specialized version script.
+/// See `RegularVersionScript` for the general case.
+pub(crate) enum VersionScript<'data> {
+    Regular(RegularVersionScript<'data>),
+    Rust(RustVersionScript<'data>),
+}
+
+impl Default for VersionScript<'_> {
+    fn default() -> Self {
+        VersionScript::Regular(RegularVersionScript::default())
+    }
+}
+
 #[derive(Debug, PartialEq, Eq, Clone)]
 pub(crate) enum SymbolMatcher<'data> {
     // Exact match.
@@ -64,7 +88,7 @@ pub(crate) enum SymbolMatcher<'data> {
     MatchesAll,
 }
 
-#[derive(Debug, Default)]
+#[derive(Debug, Default, PartialEq, Eq)]
 pub(crate) struct BasicMatchRules<'data> {
     exact: HashSet<PreHashed<UnversionedSymbolName<'data>>, PassThroughHasher>,
     escaped_exact: HashSet<Vec<u8>>,
@@ -235,7 +259,7 @@ impl<'data> SymbolLookupNameWrapper<'data> {
     }
 }
 
-impl<'data> VersionScript<'data> {
+impl<'data> RegularVersionScript<'data> {
     fn find_match(
         &self,
         name: &PreHashed<UnversionedSymbolName>,
@@ -260,8 +284,8 @@ impl<'data> VersionScript<'data> {
             }
         }
 
-        // 2) Otherwise, the last version tag with a non-* wildcard pattern wins ('global' should be checked first).
-        //    Otherwise, the last version tag with a * pattern wins.
+        // 2) Otherwise, the last version tag with a non-* wildcard pattern wins ('global' should be
+        //    checked first). Otherwise, the last version tag with a * pattern wins.
         for &non_star in &[true, false] {
             for (i, version) in self.versions.iter().enumerate().rev() {
                 let body = &version.version_body;
@@ -310,6 +334,7 @@ fn parse_version_script<'input>(input: &mut &'input BStr) -> winnow::Result<Vers
     skip_comments_and_whitespace(input)?;
 
     // Simple version script, only defines symbols visibility
+    // May be rust-style version script.
     if input.starts_with(b"{") {
         let version_body = parse_version_section(input)?;
 
@@ -317,16 +342,11 @@ fn parse_version_script<'input>(input: &mut &'input BStr) -> winnow::Result<Vers
 
         skip_comments_and_whitespace(input)?;
 
-        return Ok(VersionScript {
-            versions: vec![Version {
-                version_body,
-                ..Default::default()
-            }],
-            version_name_mapping: HashMap::new(),
-        });
+        return Ok(version_body.into());
     }
 
-    let mut version_script = VersionScript::default();
+    // Multiple versions, won't be rust-style.
+    let mut version_script = RegularVersionScript::default();
 
     // Base version placeholder
     version_names.push(b"");
@@ -337,7 +357,7 @@ fn parse_version_script<'input>(input: &mut &'input BStr) -> winnow::Result<Vers
 
         skip_comments_and_whitespace(input)?;
 
-        let version_body = parse_version_section(input)?;
+        let version_body = parse_version_section(input)?.into();
 
         let parent_name = take_until(0.., b';').parse_next(input)?;
 
@@ -373,17 +393,53 @@ fn parse_version_script<'input>(input: &mut &'input BStr) -> winnow::Result<Vers
             .insert(name, version_script.versions.len() - 1);
     }
 
-    Ok(version_script)
+    Ok(VersionScript::Regular(version_script))
 }
-
 impl<'data> VersionScript<'data> {
-    #[tracing::instrument(skip_all, name = "Parse version script")]
     pub(crate) fn parse(data: ScriptData<'data>) -> Result<VersionScript<'data>> {
+        timing_phase!("Parse version script");
+
         parse_version_script
             .parse(BStr::new(data.raw))
             .map_err(|err| error!("Failed to parse version script:\n{err}"))
     }
 
+    pub(crate) fn version_count(&self) -> u16 {
+        match self {
+            VersionScript::Regular(script) => script.version_count(),
+            VersionScript::Rust(_) => 0,
+        }
+    }
+
+    pub(crate) fn parent_count(&self) -> u16 {
+        match self {
+            VersionScript::Regular(script) => script.parent_count(),
+            VersionScript::Rust(_) => 0,
+        }
+    }
+    pub(crate) fn version_for_symbol(
+        &self,
+        name: &PreHashed<UnversionedSymbolName>,
+        version_name: Option<&[u8]>,
+    ) -> Result<Option<u16>> {
+        match self {
+            VersionScript::Regular(script) => script.version_for_symbol(name, version_name),
+            VersionScript::Rust(_) => Ok(None),
+        }
+    }
+}
+
+impl<'data> RegularVersionScript<'data> {
+    #[cfg(test)]
+    fn parse(data: ScriptData<'data>) -> Result<RegularVersionScript<'data>> {
+        match VersionScript::parse(data)? {
+            VersionScript::Regular(script) => Ok(script),
+            VersionScript::Rust(_) => {
+                bail!("Rust-style version script cannot be used as a regular version script")
+            }
+        }
+    }
+
     pub(crate) fn is_local(&self, name: &PreHashed<UnversionedSymbolName>) -> bool {
         self.find_match(name)
             .is_some_and(|(_, rule)| matches!(rule, VersionRuleSection::Local))
@@ -438,10 +494,93 @@ impl<'data> VersionScript<'data> {
     }
 }
 
-fn parse_version_section<'data>(input: &mut &'data BStr) -> winnow::Result<VersionBody<'data>> {
+#[derive(Debug, Default)]
+/// A generaic parsed version body before version script specialization optimizations.
+struct RawVersionBody<'data> {
+    pub globals: Vec<ParsedSymbolMatcher<'data>>,
+    pub locals: Vec<ParsedSymbolMatcher<'data>>,
+}
+
+impl<'data> RawVersionBody<'data> {
+    fn rust_like(&self) -> bool {
+        // one of the local has to be match-all `*` wildcard
+        if !self.locals.iter().any(|matcher| {
+            matches!(
+                matcher,
+                ParsedSymbolMatcher::Single(SymbolMatcher::MatchesAll)
+            )
+        }) {
+            return false;
+        }
+
+        // and only exact matchers in global
+        self.globals.iter().all(|matcher| {
+            matches!(
+                matcher,
+                ParsedSymbolMatcher::Single(SymbolMatcher::Exact(_))
+            )
+        })
+    }
+}
+
+impl<'data> TryFrom<RawVersionBody<'data>> for RustVersionScript<'data> {
+    type Error = RawVersionBody<'data>;
+
+    fn try_from(body: RawVersionBody<'data>) -> Result<Self, Self::Error> {
+        if !body.rust_like() {
+            return Err(body);
+        }
+        let global = body
+            .globals
+            .into_iter()
+            .map(|matcher| {
+                if let ParsedSymbolMatcher::Single(SymbolMatcher::Exact(name)) = matcher {
+                    name
+                } else {
+                    unreachable!()
+                }
+            })
+            .collect();
+
+        Ok(RustVersionScript { global })
+    }
+}
+
+impl<'data> From<RawVersionBody<'data>> for VersionScript<'data> {
+    fn from(body: RawVersionBody<'data>) -> Self {
+        match RustVersionScript::try_from(body) {
+            Ok(rust_script) => VersionScript::Rust(rust_script),
+            Err(body) => {
+                let version_body = body.into();
+                VersionScript::Regular(RegularVersionScript {
+                    versions: vec![Version {
+                        version_body,
+                        ..Default::default()
+                    }],
+                    version_name_mapping: HashMap::new(),
+                })
+            }
+        }
+    }
+}
+
+impl<'data> From<RawVersionBody<'data>> for VersionBody<'data> {
+    fn from(body: RawVersionBody<'data>) -> Self {
+        let mut out = VersionBody::default();
+        for global in body.globals {
+            out.globals.push(global);
+        }
+        for local in body.locals {
+            out.locals.push(local);
+        }
+        out
+    }
+}
+
+fn parse_version_section<'data>(input: &mut &'data BStr) -> winnow::Result<RawVersionBody<'data>> {
     let mut section = None;
 
-    let mut out = VersionBody::default();
+    let mut out = RawVersionBody::default();
 
     '{'.parse_next(input)?;
 
@@ -573,10 +712,16 @@ pub(crate) fn parse_matcher<'data>(
             let glob_type = analyze_glob_pattern(token);
 
             let create_pattern = |token: &[u8]| -> winnow::Result<Pattern> {
-                Pattern::new(str::from_utf8(token).map_err(|_| {
+                let pattern = str::from_utf8(token).map_err(|_| {
                     ContextError::from_external_error(input, VersionScriptError::InvalidUtf8String)
-                })?)
-                .map_err(|_| {
+                })?;
+                // Right now, there is a pending PR that will support the '^' as the negation
+                // character in the glob crate: https://github.com/rust-lang/glob/issues/116
+                //
+                // Let's optimistically assume the '^' cannot be part of the symbol's name (escaped
+                // in a pattern).
+                let pattern = pattern.replace("[^", "[!");
+                Pattern::new(pattern.as_str()).map_err(|_| {
                     ContextError::from_external_error(input, VersionScriptError::InvalidGlobPattern)
                 })
             };
@@ -681,13 +826,38 @@ mod tests {
     use itertools::Itertools;
     use itertools::assert_equal;
 
-    fn is_matching_global<'data>(script: &VersionScript<'data>, name: &str) -> bool {
+    fn is_matching_global<'data>(script: &RegularVersionScript<'data>, name: &str) -> bool {
         let Some(m) = script.find_match(&UnversionedSymbolName::prehashed(name.as_bytes())) else {
             return true;
         };
         matches!(m.1, VersionRuleSection::Global)
     }
 
+    #[test]
+    fn parse_rust_version_script() {
+        let data = ScriptData {
+            raw: br#"
+                    {
+                    global:
+                        foo;
+                        bar;
+                    local:
+                        *;
+                    };"#,
+        };
+        let script = VersionScript::parse(data).unwrap();
+        let VersionScript::Rust(rust_script) = script else {
+            panic!("Expected Rust-style version script");
+        };
+        assert_equal(
+            rust_script
+                .global
+                .iter()
+                .map(|sym| String::from_utf8_lossy(sym)),
+            ["foo", "bar"],
+        );
+    }
+
     #[test]
     fn test_parse_simple_version_script() {
         let data = ScriptData {
@@ -705,7 +875,7 @@ mod tests {
                         *;
                     };"#,
         };
-        let script = VersionScript::parse(data).unwrap();
+        let script = RegularVersionScript::parse(data).unwrap();
         let version_body = &script.versions[0].version_body;
         assert_equal(
             version_body
@@ -748,7 +918,7 @@ mod tests {
                 } VERS_1.1;
             "#,
         };
-        let script = VersionScript::parse(data).unwrap();
+        let script = RegularVersionScript::parse(data).unwrap();
         assert_eq!(script.versions.len(), 3);
 
         let version = &script.versions[1];
@@ -795,7 +965,7 @@ mod tests {
         let data = ScriptData {
             raw: br#"VERSION42 { global: *; };"#,
         };
-        VersionScript::parse(data).unwrap();
+        RegularVersionScript::parse(data).unwrap();
     }
 
     #[test]
@@ -814,7 +984,7 @@ mod tests {
                         };
                 };"#,
         };
-        let script = VersionScript::parse(data).unwrap();
+        let script = RegularVersionScript::parse(data).unwrap();
         let version_body = &script.versions[1].version_body;
 
         assert_equal(
@@ -875,7 +1045,7 @@ mod tests {
                         };
                 };"#,
         };
-        let script = VersionScript::parse(data).unwrap();
+        let script = RegularVersionScript::parse(data).unwrap();
         let version_body = &script.versions[1].version_body;
 
         assert_equal(
@@ -900,7 +1070,7 @@ mod tests {
                     };
                 };"#,
         };
-        let script = VersionScript::parse(data).unwrap();
+        let script = RegularVersionScript::parse(data).unwrap();
         let version_body = &script.versions[0].version_body;
 
         assert_equal(
@@ -922,7 +1092,7 @@ mod tests {
                     };
                 };"#,
         };
-        let script = VersionScript::parse(data).unwrap();
+        let script = RegularVersionScript::parse(data).unwrap();
         let version_body = &script.versions[0].version_body;
 
         assert_equal(
@@ -977,7 +1147,7 @@ mod tests {
                 } VERS_1.1;
             "#,
         };
-        let script = VersionScript::parse(data).unwrap();
+        let script = RegularVersionScript::parse(data).unwrap();
         let sym = UnversionedSymbolName::prehashed;
 
         // Exact match wins
@@ -1006,7 +1176,7 @@ mod tests {
                 };
             "#,
         };
-        let script = VersionScript::parse(data).unwrap();
+        let script = RegularVersionScript::parse(data).unwrap();
         let version_body = &script.versions[0].version_body;
 
         let escaped_patterns: HashSet<&[u8]> = version_body
@@ -1042,4 +1212,24 @@ mod tests {
 
         assert!(nonstar_patterns.contains(&"f?"));
     }
+
+    #[test]
+    fn test_negation() {
+        let data = ScriptData {
+            raw: br#"
+                VERS_1.1 {
+                    f*;
+                };
+
+                VERS_1.2 {
+                    f[^o]*;
+                } VERS_1.1;
+            "#,
+        };
+        let script = RegularVersionScript::parse(data).unwrap();
+        let sym = UnversionedSymbolName::prehashed;
+
+        assert_eq!(script.find_match(&sym(b"foo")).unwrap().0, 1);
+        assert_eq!(script.find_match(&sym(b"fxxx")).unwrap().0, 2);
+    }
 }
