@@ -3,7 +3,7 @@
 //! work unless they're performed. e.g. it uses GOT relocations in _start, which cannot work in a
 //! static-PIE binary because dynamic relocations haven't yet been applied to the GOT yet.
 
-use crate::args::OutputKind;
+use crate::OutputKind;
 use crate::elf::PLT_ENTRY_SIZE;
 use crate::error;
 use crate::error::Result;
@@ -43,6 +43,8 @@ const _ASSERTS: () = {
 impl crate::arch::Arch for X86_64 {
     type Relaxation = Relaxation;
 
+    const KIND: crate::arch::Architecture = crate::arch::Architecture::X86_64;
+
     fn elf_header_arch_magic() -> u16 {
         object::elf::EM_X86_64
     }
@@ -106,6 +108,10 @@ impl crate::arch::Arch for X86_64 {
     fn merge_eflags(_eflags: &[u32]) -> Result<u32> {
         Ok(0)
     }
+
+    fn high_part_relocations() -> &'static [u32] {
+        &[]
+    }
 }
 
 macro_rules! rel_info_from_type {
@@ -363,7 +369,8 @@ impl crate::arch::Relaxation for Relaxation {
             object::elf::R_X86_64_GOTPC32_TLSDESC
                 if !interposable && output_kind.is_executable() =>
             {
-                // We require that the instruction that this relocation applies to is a LEA instruction.
+                // We require that the instruction that this relocation applies to is a LEA
+                // instruction.
                 let bytes = section_bytes.get(offset - 3..offset - 1);
                 if bytes == Some(&[0x48, 0x8d]) || bytes == Some(&[0x4c, 0x8d]) {
                     return Some(Relaxation {
@@ -376,7 +383,8 @@ impl crate::arch::Relaxation for Relaxation {
             // Note, the conditions on this relaxation (is_executable) must match those on
             // TLSDESC_CALL below.
             object::elf::R_X86_64_GOTPC32_TLSDESC if output_kind.is_executable() => {
-                // We require that the instruction that this relocation applies to is a LEA instruction.
+                // We require that the instruction that this relocation applies to is a LEA
+                // instruction.
                 let bytes = section_bytes.get(offset - 3..offset - 1);
                 if bytes == Some(&[0x48, 0x8d]) || bytes == Some(&[0x4c, 0x8d]) {
                     return Some(Relaxation {
