@@ -42,11 +42,13 @@ mod eh_frame_diff;
 mod gnu_hash;
 mod header_diff;
 mod init_order;
+mod loongarch64;
 mod riscv64;
 pub(crate) mod section_map;
 mod segment;
 mod symbol_diff;
 mod symtab;
+mod sysv_hash;
 mod trace;
 mod utils;
 mod version_diff;
@@ -143,15 +145,13 @@ impl Config {
     fn apply_wild_defaults(&mut self, arch: ArchKind) {
         self.ignore.extend(
             [
-                // We don't currently support allocating space except in sections, so we have sections
-                // to hold the section and program headers. We then need to ignore them because GNU ld
-                // doesn't define such sections.
+                // We don't currently support allocating space except in sections, so we have
+                // sections to hold the section and program headers. We then need
+                // to ignore them because GNU ld doesn't define such sections.
                 "section.shdr",
                 "section.phdr",
                 // We don't yet support these sections.
                 "section.data.rel.ro",
-                "section.hash",
-                "section.sframe",
                 // We set this to 8. GNU ld sometimes does too, but sometimes to 0.
                 "section.got.entsize",
                 "section.plt.got.entsize",
@@ -164,9 +164,9 @@ impl Config {
                 "section.annobin.notes.entsize",
                 // We don't yet group .lrodata sections separately.
                 "section.lrodata",
-                // We sometimes eliminate __tls_get_addr where GNU ld doesn't. This can mean that we
-                // have no versioned symbols for ld-linux-x86-64.so.2 or equivalent, which means we
-                // end up with one less version record.
+                // We sometimes eliminate __tls_get_addr where GNU ld doesn't. This can mean that
+                // we have no versioned symbols for ld-linux-x86-64.so.2 or
+                // equivalent, which means we end up with one less version record.
                 ".dynamic.DT_VERNEEDNUM",
                 // We currently handle these dynamic tags differently
                 ".dynamic.DT_JMPREL",
@@ -177,22 +177,28 @@ impl Config {
                 GOT_PLT_SECTION_NAME_STR,
                 // We don't currently produce a separate .plt.sec section.
                 "section.plt.sec",
-                // We don't yet write this.
+                // Different hash values due to different implementations.
                 ".dynamic.DT_HASH",
-                // aarch64-linux-gnu-ld on arch linux emits DT_BIND_NOW instead of DT_FLAGS.BIND_NOW
+                // Different hash values due to different implementations.
+                ".hash",
+                "section.hash.alignment",
+                "section.hash.entsize",
+                // Some other linkers seem to generate a `.hash` section even when there are no
+                // dynamic symbols.
+                "section.hash",
+                // aarch64-linux-gnu-ld on arch linux emits DT_BIND_NOW instead of
+                // DT_FLAGS.BIND_NOW
                 ".dynamic.DT_BIND_NOW",
                 ".dynamic.DT_FLAGS.BIND_NOW",
-                // TODO: Implement proper ordering of .init .ctors etc
-                "init_array",
-                "fini_array",
                 // When GNU ld encounters a GOT-forming reference to an ifunc, it generates a
                 // canonical PLT entry and points the GOT at that. This means that it ends up with
                 // GOT->PLT->GOT. We don't as yet support doing this.
                 "rel.missing-got-plt-got",
-                // We do support this. TODO: Should definitely look into why we're seeing this missing
-                // in our output.
+                // We do support this. TODO: Should definitely look into why we're seeing this
+                // missing in our output.
                 "section.rela.plt",
-                // We currently write 10 byte PLT entries in some cases where GNU ld writes 8 byte ones.
+                // We currently write 10 byte PLT entries in some cases where GNU ld writes 8 byte
+                // ones.
                 "section.plt.got.alignment",
                 // GNU ld sometimes makes this writable sometimes not. Presumably this depends on
                 // whether there are relocations or some flags.
@@ -227,8 +233,8 @@ impl Config {
                 // `-zdynamic-undefined-weak`.
                 "rel.undefined-weak.dynamic.R_X86_64_64",
                 "rel.undefined-weak.dynamic.R_AARCH64_ABS64",
-                // On aarch64, GNU ld, at least sometimes, converts R_AARCH64_ABS64 to a PLT-forming
-                // relocation. We at present, don't.
+                // On aarch64, GNU ld, at least sometimes, converts R_AARCH64_ABS64 to a
+                // PLT-forming relocation. We at present, don't.
                 "rel.dynamic-plt-bypass",
                 // If we don't optimise a TLS access, then we'll have references to __tls_get_addr,
                 // when GNU ld doesn't.
@@ -239,13 +245,19 @@ impl Config {
                 "segment.PHDR.*",
                 "segment.GNU_RELRO.*",
                 "segment.GNU_STACK.*",
-                "segment.GNU_PROPERTY.*",
-                "segment.GNU_SFRAME.*",
                 // Wild currently generates PT_NOTE even for non-alloc note sections, while the
                 // other linkers don't.
                 "segment.NOTE.*",
                 // TODO: RISC-V
                 "segment.LOAD.RW.alignment",
+                // TODO: Latest lld sometimes doesn’t create a .note.gnu.property section even when
+                // Wild does.
+                "segment.GNU_PROPERTY.alignment",
+                "segment.GNU_PROPERTY.flags",
+                // GNU ld and lld sometimes don’t generate .sframe sections in cases where we do.
+                // TODO: Figure out why this is happening.
+                "segment.GNU_SFRAME.alignment",
+                "segment.GNU_SFRAME.flags",
             ]
             .into_iter()
             .map(ToOwned::to_owned),
@@ -254,8 +266,8 @@ impl Config {
         match arch {
             ArchKind::Aarch64 => self.ignore.extend(
                 [
-                    // Other linkers have a bigger initial PLT entry, thus the entsize is set to zero:
-                    // https://sourceware.org/bugzilla/show_bug.cgi?id=26312
+                    // Other linkers have a bigger initial PLT entry, thus the entsize is set to
+                    // zero: https://sourceware.org/bugzilla/show_bug.cgi?id=26312
                     "section.plt.entsize",
                     // On Alpine Linux, aarch64, GNU ld seems to emit the _DYNAMIC symbol without a
                     // section index instead of pointing it at the .dynamic section.
@@ -286,6 +298,18 @@ impl Config {
                 .map(ToOwned::to_owned),
             ),
             ArchKind::X86_64 => {}
+            ArchKind::LoongArch64 => self.ignore.extend(
+                [
+                    "section.sdata",
+                    "section.iplt",
+                    "rel.unknown_failure*",
+                    "literal-byte-mismatch*",
+                    "error.*",
+                    "section-diff-failed*",
+                ]
+                .into_iter()
+                .map(ToOwned::to_owned),
+            ),
         }
 
         self.equiv.push((
@@ -598,6 +622,12 @@ impl Report {
             GNU_HASH_SECTION_NAME_STR,
             gnu_hash::check_object,
         );
+        validate_objects(
+            self,
+            objects,
+            HASH_SECTION_NAME_STR,
+            sysv_hash::check_object,
+        );
         validate_objects(self, objects, "index", asm_diff::validate_indexes);
         validate_objects(
             self,
@@ -637,6 +667,9 @@ impl Report {
             ArchKind::RISCV64 => {
                 self.report_arch_specific_diffs::<crate::riscv64::RiscV64>(objects);
             }
+            ArchKind::LoongArch64 => {
+                self.report_arch_specific_diffs::<crate::loongarch64::LoongArch64>(objects);
+            }
         }
     }
 
