@@ -194,12 +194,12 @@ impl Arch for X86_64 {
         relaxation_kind.apply(section_bytes, offset_in_section, addend);
     }
 
-    fn decode_instructions_in_range<'data>(
-        section_bytes: &'data [u8],
+    fn decode_instructions_in_range(
+        section_bytes: &[u8],
         section_address: u64,
         function_offset_in_section: u64,
         range: std::ops::Range<u64>,
-    ) -> Vec<Instruction<'data, Self>> {
+    ) -> Vec<Instruction<'_, Self>> {
         let mut instructions = Vec::new();
 
         let mut decoder = AsmDecoder::new(
@@ -284,9 +284,9 @@ impl Arch for X86_64 {
                     0xf2, 0xe9, 0, 0, 0, 0,    // bnd jmp {plt[0]}(%rip)
                     0x90, // nop
                 ];
-                // Note: Some variants use jmp instead of bnd jmp, then a different padding instruction.
-                // Because we use the index that gets pushed, we ignore the bytes of the later
-                // instructions, so that we support these variants.
+                // Note: Some variants use jmp instead of bnd jmp, then a different padding
+                // instruction. Because we use the index that gets pushed, we ignore the bytes of
+                // the later instructions, so that we support these variants.
                 if plt_entry[..5] == PLT_ENTRY_TEMPLATE[..5] {
                     let index = u32_from_slice(&plt_entry[5..]);
                     return Some(PltEntry::JumpSlot(index));
